"""
PC-BASIC - implementation.py
Top-level implementation and main interpreter loop

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""
import io
import os
import sys
import math
import logging
from functools import partial
from contextlib import contextmanager

from ..compat import queue, text_type

from .data import NAME, VERSION, COPYRIGHT
from .base import error
from .base import tokens as tk
from .base import signals
from .base import codestream
from .devices import Files, InputTextFile
from . import converter
from . import eventcycle
from . import basicevents
from . import program
from . import display
from . import console
from . import inputs
from . import clock
from . import dos
from . import memory
from . import machine
from . import interpreter
from . import sound
from . import iostreams
from . import codepage as cp
from . import values
from . import parser
from . import extensions


GREETING = (
    b'KEY ON:PRINT "%s %s":PRINT "%s":PRINT USING "##### Bytes free"; FRE(0)'
    % tuple(s.encode('ascii') for s in (NAME, VERSION, COPYRIGHT))
)


class Implementation(object):
    """Interpreter session, implementation class."""

    def __init__(
            self, syntax=u'advanced', double=False, term=u'', shell=u'',
            output_streams=u'stdio', input_streams=u'stdio',
            codepage=None, box_protect=True, font=None, text_width=80,
            video=u'cga', monitor=u'rgb',
            devices=None, current_device=u'Z:',
            textfile_encoding=None, soft_linefeed=False,
            check_keybuffer_full=True, ctrl_c_is_break=True,
            hide_listing=None, hide_protected=False,
            peek_values=None, allow_code_poke=False, rebuild_offsets=True,
            max_memory=65534, reserved_memory=3429, video_memory=262144,
            serial_buffer_size=128, max_reclen=128, max_files=3,
            extension=()
        ):
        """Initialise the interpreter session."""
        ######################################################################
        # session-level members
        ######################################################################
        # true if a prompt is needed on next cycle
        self._prompt = True
        # AUTO mode state
        self._auto_mode = False
        self._auto_linenum = 10
        self._auto_increment = 10
        # syntax error prompt and EDIT
        self._edit_prompt = False
        # terminal program for TERM command
        self._term_program = term
        ######################################################################
        # data segment
        ######################################################################
        # set up variables and memory model state
        # initialise the data segment
        self.memory = memory.DataSegment(
            max_memory, reserved_memory, max_reclen, max_files, double
        )
        # values and variables
        self.strings = self.memory.strings
        self.values = self.memory.values
        self.scalars = self.memory.scalars
        self.arrays = self.memory.arrays
        # prepare tokeniser
        token_keyword = tk.TokenKeywordDict(syntax)
        self.tokeniser = converter.Tokeniser(self.values, token_keyword)
        self.lister = converter.Lister(self.values, token_keyword)
        # initialise the program
        bytecode = codestream.TokenisedStream(self.memory.code_start)
        self.program = program.Program(
            self.tokeniser, self.lister, hide_listing, hide_protected,
            allow_code_poke, self.memory, bytecode, rebuild_offsets
        )
        # register all data segment users
        self.memory.set_buffers(self.program)
        ######################################################################
        # console
        ######################################################################
        # prepare codepage
        self.codepage = cp.Codepage(codepage, box_protect)
        # set up input event handler
        # no interface yet; use dummy queues
        self.queues = eventcycle.EventQueues(ctrl_c_is_break, inputs=queue.Queue())
        # prepare I/O streams
        self.io_streams = iostreams.IOStreams(self.queues, self.codepage)
        self.io_streams.add_pipes(input=input_streams)
        self.io_streams.add_pipes(output=output_streams)
        # initialise sound queue
        self.sound = sound.Sound(self.queues, self.values, self.memory, syntax)
        # initialise video
        self.display = display.Display(
            self.queues, self.values, self.queues,
            self.memory, text_width, video_memory, video, monitor,
            self.codepage, font
        )
        self.text_screen = self.display.text_screen
        self.graphics = self.display.graphics
        # prepare input devices (keyboard, pen, joystick, clipboard-copier)
        # EventHandler needed for wait() only
        self.keyboard = inputs.Keyboard(
            self.queues, self.values, self.codepage, check_keybuffer_full
        )
        self.pen = inputs.Pen()
        self.stick = inputs.Stick(self.values)
        # 12 definable function keys for Tandy, BASICA, PCjr
        # regular GW-BASIC should only have 10
        num_fn_keys = 10 if syntax == 'gwbasic' else 12
        tandy_fn_keys = syntax == 'tandy'
        # initialise the console
        # Sound is needed for the beeps on \a
        self.console = console.Console(
            self.text_screen, self.display.cursor,
            self.keyboard, self.sound, self.io_streams, num_fn_keys, tandy_fn_keys
        )
        # initilise floating-point error message stream
        self.values.set_handler(values.FloatErrorHandler(self.console))
        ######################################################################
        # devices
        ######################################################################
        # intialise devices and files
        # DataSegment needed for COMn and disk FIELD buffers
        # EventCycle needed for wait()
        self.files = Files(
            self.values, self.memory, self.queues, self.keyboard, self.display, self.console,
            max_files, max_reclen, serial_buffer_size,
            devices, current_device,
            self.codepage, textfile_encoding, soft_linefeed
        )
        # enable printer echo from console
        self.console.set_lpt1_file(self.files.lpt1_file)
        ######################################################################
        # other components
        ######################################################################
        # set up the SHELL command
        # Files needed for current disk device
        self.shell = dos.Shell(
            self.queues, self.keyboard, self.console, self.files, self.codepage, shell
        )
        # set up environment
        self.environment = dos.Environment(self.values, self.codepage)
        # initialise random number generator
        self.randomiser = values.Randomiser(self.values)
        # initialise system clock
        self.clock = clock.Clock(self.values)
        ######################################################################
        # register input event handlers
        ######################################################################
        # clipboard and print screen handler
        self.queues.add_handler(display.ScreenCopyHandler(
            self.queues, self.text_screen, self.files.lpt1_file
        ))
        # keyboard, pen and stick
        self.queues.add_handler(self.keyboard)
        self.queues.add_handler(self.pen)
        self.queues.add_handler(self.stick)
        # set up BASIC event handlers
        self.basic_events = basicevents.BasicEvents(
            self.sound, self.clock, self.files, self.program, num_fn_keys, tandy_fn_keys
        )
        ######################################################################
        # extensions
        ######################################################################
        self.extensions = extensions.Extensions(extension, self.values, self.codepage)
        ######################################################################
        # interpreter
        ######################################################################
        # initialise the parser
        self.parser = parser.Parser(self.values, self.memory, syntax)
        # initialise the interpreter
        self.interpreter = interpreter.Interpreter(
            self.queues, self.console, self.display.cursor, self.files, self.sound,
            self.values, self.memory, self.program, self.parser, self.basic_events
        )
        ######################################################################
        # callbacks
        ######################################################################
        # set up non-data segment memory
        self.all_memory = machine.Memory(
            self.values, self.memory, self.files,
            self.display, self.keyboard, self.display.memory_font,
            self.interpreter, peek_values, syntax
        )
        # initialise machine ports
        self.machine = machine.MachinePorts(
            self.queues, self.values, self.display, self.keyboard, self.stick, self.files
        )
        # build function table (depends on Memory having been initialised)
        self.parser.init_callbacks(self)

    def __getstate__(self):
        """Pickle the session."""
        return self.__dict__

    def __setstate__(self, pickle_dict):
        """Unpickle and resume the session."""
        self.__dict__.update(pickle_dict)
        # re-assign callbacks (not picklable)
        self.parser.init_callbacks(self)
        # reopen keyboard, in case we quit because it was closed
        self.keyboard._input_closed = False
        # suppress double prompt
        if not self.interpreter.parse_mode:
            self._prompt = False

    def attach_interface(self, interface=None):
        """Attach interface to interpreter session."""
        if interface:
            self.queues.set(*interface.get_queues())
            # rebuild the screen
            self.display.rebuild()
            # rebuild audio queues
            self.sound.rebuild()
        else:
            # use dummy video & audio queues if not provided
            # but an input queue should be operational for I/O streams
            self.queues.set(inputs=queue.Queue())

    def execute(self, command):
        """Execute a BASIC statement."""
        with self._handle_exceptions():
            self._store_line(command)
            self.interpreter.loop()

    def evaluate(self, expression):
        """Evaluate a BASIC expression."""
        with self._handle_exceptions():
            # prefix expression with a PRINT token
            # to avoid any number at the start to be taken as a line number
            tokens = self.tokeniser.tokenise_line(b'?' + expression)
            # skip : and ? tokens and parse expression
            tokens.read(2)
            val =  self.parser.parse_expression(tokens)
            return val.to_value()
        return None

    def set_variable(self, name, value):
        """Set a variable in memory."""
        name = name.upper()
        if isinstance(value, text_type):
            value = self.codepage.unicode_to_bytes(value)
        elif isinstance(value, bool):
            value = -1 if value else 0
        if b'(' in name:
            name = name.split(b'(', 1)[0]
            self.arrays.from_list(value, name)
        else:
            self.memory.set_variable(name, [], self.values.from_value(value, name[-1:]))

    def get_converter(self, from_type, to_type):
        """Get a converter function; raise ValueError if not allowed"""
        if to_type is None or from_type == to_type:
            return lambda _x: _x
        converter = {
            (bytes, text_type): partial(self.codepage.bytes_to_unicode, preserve=cp.CONTROL),
            (text_type, bytes): self.codepage.unicode_to_bytes,
            (int, bool): bool,
            (float, bool): bool,
            (bool, int): lambda _bool: (-1 if _bool else 0),
            (int, float): float,
            (float, int): lambda _flt: int(math.floor(_flt)),
            (bool, float): lambda _bool: (-1. if _bool else 0.),
        }
        try:
            return converter[(from_type, to_type)]
        except KeyError:
            raise ValueError("BASIC can't convert %s to %s." % (from_type, to_type))

    def get_variable(self, name, as_type=None):
        """Get a variable in memory."""
        name = name.upper()
        if b'(' in name:
            name = name.split(b'(', 1)[0]
            value = self.arrays.to_list(name)
            if not value:
                return []
            convert = self.get_converter(type(value[0]), as_type)
            return [convert(_item) for _item in value]
        else:
            value = self.memory.view_or_create_variable(name, []).to_value()
            convert = self.get_converter(type(value), as_type)
            return convert(value)

    def interact(self):
        """Interactive interpreter session."""
        while True:
            with self._handle_exceptions():
                self.interpreter.loop()
                if self._auto_mode:
                    self._auto_step()
                else:
                    self._show_prompt()
                    # input loop, checks events
                    line = self.console.read_line(is_input=False)
                    self._prompt = not self._store_line(line)

    def close(self):
        """Close the session."""
        # close files if we opened any
        self.files.close_all()
        self.files.close_devices()
        # kill the iostreams threads so windows doesn't run out
        self.io_streams.close()

    def _show_prompt(self):
        """Show the Ok or EDIT prompt, unless suppressed."""
        if self._prompt:
            self.console.start_line()
            self.console.write_line(b'Ok\xff')
        if self._edit_prompt:
            linenum, tell = self._edit_prompt
            # unset edit prompt first, in case program.edit throws
            self._edit_prompt = False
            self.program.edit(self.console, linenum, tell)

    def _store_line(self, line):
        """Store a program line or schedule a command line for execution."""
        if not line:
            return True
        self.interpreter.direct_line = self.tokeniser.tokenise_line(line)
        c = self.interpreter.direct_line.peek()
        if c == b'\0':
            # clear all program stacks
            self.interpreter.clear_stacks_and_pointers()
            # clear variables first,
            # to avoid inconsistent state in string space if out of memory
            self._clear_all()
            # check for lines starting with numbers (6553 6) and empty lines
            self.program.check_number_start(self.interpreter.direct_line)
            self.program.store_line(self.interpreter.direct_line)
            return True
        elif c != b'':
            self.interpreter.run_mode = False
            # it is a command, go and execute
            self.interpreter.set_parse_mode(True)
            return False

    def _auto_step(self):
        """Generate an AUTO line number and wait for input."""
        try:
            numstr = b'%d' % (self._auto_linenum,)
            if self._auto_linenum in self.program.line_numbers:
                prompt = numstr + b'*'
            else:
                prompt = numstr + b' '
            line = self.console.read_line(prompt, is_input=False)
            # remove *, if present
            if line[:len(numstr)+1] == b'%s*' % (numstr,):
                line = b'%s %s' % (numstr, line[len(numstr)+1:])
            # run or store it; don't clear lines or raise undefined line number
            self.interpreter.direct_line = self.tokeniser.tokenise_line(line)
            c = self.interpreter.direct_line.peek()
            if c == b'\0':
                # check for lines starting with numbers (6553 6) and empty lines
                empty, scanline = self.program.check_number_start(self.interpreter.direct_line)
                if not empty:
                    self.program.store_line(self.interpreter.direct_line)
                    # clear all program stacks
                    self.interpreter.clear_stacks_and_pointers()
                    self._clear_all()
                self._auto_linenum = scanline + self._auto_increment
            elif c != b'':
                # it is a command, go and execute
                self.interpreter.set_parse_mode(True)
        except error.Break:
            # ctrl+break, ctrl-c both stop background sound
            self.sound.stop_all_sound()
            self._auto_mode = False


    ##############################################################################
    # error handling

    @contextmanager
    def _handle_exceptions(self):
        """Context guard to handle BASIC exceptions."""
        try:
            yield
        except error.Break as e:
            # ctrl-break stops foreground and background sound
            self.sound.stop_all_sound()
            if not self.interpreter.parse_mode:
                self._prompt = False
            else:
                self.interpreter.set_pointer(False)
                # call _handle_error to write a message, etc.
                self._handle_error(e)
                # override position of syntax error
                if e.trapped_error_num == error.STX:
                    self._syntax_error_edit_prompt(e.trapped_error_pos)
        except error.BASICError as e:
            self._handle_error(e)
        except error.Exit:
            raise

    def _handle_error(self, e):
        """Handle a BASIC error through error message."""
        # not handled by ON ERROR, stop execution
        self.console.start_line()
        self.console.write(e.get_message(self.program.get_line_number(e.pos)))
        if not self.interpreter.input_mode:
            self.console.write(b'\xFF')
        self.console.write(b'\r')
        self.interpreter.set_parse_mode(False)
        self.interpreter.input_mode = False
        self._prompt = True
        # special case: syntax error
        if e.err == error.STX:
            self._syntax_error_edit_prompt(e.pos)

    def _syntax_error_edit_prompt(self, pos):
        """Show an EDIT prompt at the location of a syntax error."""
        # for some reason, err is reset to zero by GW-BASIC in this case.
        self.interpreter.error_num = 0
        if pos is not None and pos != -1:
            # line edit gadget appears
            self._edit_prompt = (self.program.get_line_number(pos), pos+1)

    ###########################################################################
    # callbacks

    def system_(self, args):
        """SYSTEM: exit interpreter."""
        list(args)
        raise error.Exit()

    def clear_(self, args):
        """CLEAR: clear memory and redefine memory limits."""
        try:
            # positive integer expression allowed but not used
            intexp = next(args)
            if intexp is not None:
                expr = values.to_int(intexp)
                error.throw_if(expr < 0)
            # set size of BASIC memory
            mem_size = next(args)
            if mem_size is not None:
                mem_size = values.to_int(mem_size, unsigned=True)
                self.memory.set_basic_memory_size(mem_size)
            # set aside stack space for GW-BASIC.
            # default is the previous stack space size.
            stack_size = next(args)
            if stack_size is not None:
                stack_size = values.to_int(stack_size, unsigned=True)
                self.memory.set_stack_size(stack_size)
            # select video memory size (Tandy/PCjr only)
            video_size = next(args)
            if video_size is not None:
                video_size = round(video_size.to_value())
                self.display.set_video_memory_size(video_size)
            # execute any remaining parsing steps
            next(args)
        except StopIteration:
            pass
        self._clear_all()

    def _clear_all(self, close_files=False,
              preserve_functions=False, preserve_base=False, preserve_deftype=False):
        """Clear everything required for the CLEAR command."""
        if close_files:
            # close all files
            self.files.close_all()
        # Resets the stack and string space
        # Clears all COMMON and user variables
        # release all disk buffers (FIELD)?
        self.memory.clear(preserve_base, preserve_deftype)
        if not preserve_functions:
            self.parser.user_functions.clear()
        # Resets STRIG to off
        self.stick.is_on = False
        # stop all sound
        self.sound.stop_all_sound()
        # reset PLAY state
        self.sound.reset_play()
        # reset DRAW state (angle, scale) and current graphics position
        self.graphics.reset()
        # reset random number generator
        self.randomiser.clear()
        # reset stacks & pointers
        self.interpreter.clear()

    def shell_(self, args):
        """SHELL: open OS shell and optionally execute command."""
        cmd = values.next_string(args)
        list(args)
        # force cursor visible
        self.display.cursor.set_override(True)
        # sound stops playing and is forgotten
        self.sound.stop_all_sound()
        # run the os-specific shell
        self.shell.launch(cmd)
        # reset cursor visibility to its previous state
        self.display.cursor.set_override(False)

    def term_(self, args):
        """TERM: terminal emulator."""
        list(args)
        self._clear_all()
        self.interpreter.tron = False
        if not self._term_program:
            # on Tandy, raises Internal Error
            # and deletes the program currently in memory
            raise error.BASICError(error.INTERNAL_ERROR)
        # terminal program for TERM command
        prog = self.files.get_device(b'@:').bind(self._term_program)
        with self.files.open(0, prog, filetype=b'ABP', mode=b'I') as progfile:
            self.program.load(progfile)
        self.interpreter.error_handle_mode = False
        self.interpreter.clear_stacks_and_pointers()
        self.interpreter.set_pointer(True, 0)

    def delete_(self, args):
        """DELETE: delete range of lines from program."""
        line_range, = args
        # throws back to direct mode
        self.program.delete(*line_range)
        # clear all program stacks
        self.interpreter.clear_stacks_and_pointers()
        # clear all variables
        self._clear_all()

    def list_(self, args):
        """LIST: output program lines."""
        line_range = next(args)
        out = values.next_string(args)
        if out is not None:
            out = self.files.open(0, out, filetype=b'A', mode=b'O')
        list(args)
        lines = self.program.list_lines(*line_range)
        if out:
            with out:
                for l in lines:
                    out.write_line(l)
        else:
            for l in lines:
                # flow of listing is visible on screen
                # and interruptible
                self.queues.wait()
                # LIST on screen is slightly different from just writing
                self.console.list_line(l, newline=True)
        # return to direct mode
        self.interpreter.set_pointer(False)

    def edit_(self, args):
        """EDIT: output a program line and position cursor for editing."""
        from_line, = args
        from_line, = self.program.explicit_lines(from_line)
        self.program.last_stored = from_line
        if from_line is None or from_line not in self.program.line_numbers:
            raise error.BASICError(error.UNDEFINED_LINE_NUMBER)
        # throws back to direct mode
        # jump to end of direct line so execution stops
        self.interpreter.set_pointer(False)
        # request edit prompt but no Ok prompt
        self._prompt = False
        self._edit_prompt = (from_line, None)

    def auto_(self, args):
        """AUTO: enter automatic line numbering mode."""
        linenum, increment = args
        from_line, = self.program.explicit_lines(linenum)
        # reset linenum and increment on each call of AUTO (even in AUTO mode)
        self._auto_linenum = linenum if linenum is not None else 10
        self._auto_increment = increment if increment is not None else 10
        # move program pointer to end
        self.interpreter.set_pointer(False)
        # continue input in AUTO mode
        self._auto_mode = True

    def load_(self, args):
        """LOAD: load program from file."""
        name = values.next_string(args)
        comma_r, = args
        # clear variables, stacks & pointers
        self._clear_all()
        with self.files.open(0, name, filetype=b'ABP', mode=b'I') as f:
            self.program.load(f)
        # reset stacks
        self.interpreter.clear_stacks_and_pointers()
        if comma_r:
            # in ,R mode, don't close files; run the program
            self.interpreter.set_pointer(True, 0)
        else:
            self.files.close_all()
        self.interpreter.tron = False

    def chain_(self, args):
        """CHAIN: load program and chain execution."""
        merge = next(args)
        name = values.next_string(args)
        jumpnum = next(args)
        if jumpnum is not None:
            jumpnum = values.to_int(jumpnum, unsigned=True)
        preserve_all, delete_lines = next(args), next(args)
        from_line, to_line = delete_lines if delete_lines else (None, None)
        if to_line is not None and to_line not in self.program.line_numbers:
            raise error.BASICError(error.IFC)
        list(args)
        if self.program.protected and merge:
            raise error.BASICError(error.IFC)
        # gather COMMON declarations
        common_scalars, common_arrays = self.interpreter.gather_commons()
        with self.memory.preserve_commons(common_scalars, common_arrays, preserve_all):
            # preserve DEFtype on MERGE
            # functions are cleared except when CHAIN ... ALL is specified
            # OPTION BASE is preserved when there are common variables
            self._clear_all(
                    preserve_functions=preserve_all,
                    preserve_base=(common_scalars or common_arrays or preserve_all),
                    preserve_deftype=merge)
            # load new program
            with self.files.open(0, name, filetype=b'ABP', mode=b'I') as f:
                if delete_lines:
                    # delete lines from existing code before merge
                    # (without MERGE, this is pointless)
                    self.program.delete(*delete_lines)
                if merge:
                    self.program.merge(f)
                else:
                    self.program.load(f)
                # clear all program stacks
                self.interpreter.clear_stacks_and_pointers()
                # don't close files!
                # RUN
                self.interpreter.jump(jumpnum, err=error.IFC)
        # ensure newly allocated strings are not considered temporary
        # e.g. code strings in the old program become allocated strings in the new
        self.strings.fix_temporaries()

    def save_(self, args):
        """SAVE: save program to a file."""
        name = values.next_string(args)
        mode = (next(args) or b'B').upper()
        list(args)
        with self.files.open(
                0, name, filetype=mode, mode=b'O',
                seg=self.memory.data_segment, offset=self.memory.code_start,
                length=len(self.program.bytecode.getvalue())-1
            ) as f:
            self.program.save(f)
        if mode == b'A':
            # return to direct mode
            self.interpreter.set_pointer(False)

    def merge_(self, args):
        """MERGE: merge lines from file into current program."""
        name = values.next_string(args)
        list(args)
        # check if file exists, make some guesses (all uppercase, +.BAS) if not
        with self.files.open(0, name, filetype=b'A', mode=b'I') as f:
            self.program.merge(f)
        # clear all program stacks
        self.interpreter.clear_stacks_and_pointers()

    def new_(self, args):
        """NEW: clear program from memory."""
        list(args)
        self.interpreter.tron = False
        # deletes the program currently in memory
        self.program.erase()
        # reset stacks
        self.interpreter.clear_stacks_and_pointers()
        # and clears all variables
        self._clear_all()
        self.interpreter.set_pointer(False)

    def run_(self, args):
        """RUN: start program execution."""
        jumpnum = next(args)
        comma_r = False
        if jumpnum is None:
            try:
                name = values.next_string(args)
                comma_r = next(args)
                with self.files.open(0, name, filetype=b'ABP', mode=b'I') as f:
                    self.program.load(f)
            except StopIteration:
                pass
        list(args)
        self.interpreter.on_error = 0
        self.interpreter.error_handle_mode = False
        self.interpreter.clear_stacks_and_pointers()
        self._clear_all(close_files=not comma_r)
        if jumpnum is None:
            self.interpreter.set_pointer(True, 0)
        else:
            if jumpnum not in self.program.line_numbers:
                raise error.BASICError(error.UNDEFINED_LINE_NUMBER)
            self.interpreter.jump(jumpnum)

    def end_(self, args):
        """END: end program execution and return to interpreter."""
        list(args)
        # enable CONT
        self.program.bytecode.skip_to(tk.END_STATEMENT)
        self.interpreter.stop_pos = self.program.bytecode.tell()
        # jump to end of direct line so execution stops
        self.interpreter.set_pointer(False)
        # avoid NO RESUME
        self.interpreter.error_handle_mode = False
        self.interpreter.error_resume = None
        self.files.close_all()

    def input_(self, args):
        """INPUT: request input from user or read from file."""
        file_number = next(args)
        if file_number is not None:
            file_number = values.to_int(file_number)
            error.range_check(0, 255, file_number)
            finp = self.files.get(file_number, mode=b'IR')
            self._input_file(finp, args)
        else:
            newline, prompt, following = next(args)
            self._input_console(newline, prompt, following, args)

    def _input_console(self, newline, prompt, following, readvar):
        """INPUT: request input from user."""
        if following == b';':
            prompt += b'? '
        with self.memory.get_stack() as stack:
            # read the input
            self.interpreter.input_mode = True
            self.parser.redo_on_break = True
            # readvar is a list of (name, indices) tuples
            # we return a list of (name, indices, values) tuples
            while True:
                line = self.console.read_line(prompt, write_endl=newline, is_input=True)
                inputstream = InputTextFile(line)
                # read the values and group them and the separators
                var, values, seps = [], [], []
                for name, indices in readvar:
                    name = self.memory.complete_name(name)
                    word, sep = inputstream.input_entry(
                        name[-1:], allow_past_end=True, suppress_unquoted_linefeed=False
                    )
                    try:
                        value = self.values.from_repr(word, allow_nonnum=False, typechar=name[-1:])
                    except error.BASICError as e:
                        # string entered into numeric field
                        value = None
                    stack.append(value)
                    var.append([name, indices])
                    values.append(value)
                    seps.append(sep)
                # last separator not empty: there were too many values or commas
                # earlier separators empty: there were too few values
                # empty values will be converted to zero by from_str
                # None means a conversion error occurred
                if (seps[-1] or b'' in seps[:-1] or None in values):
                    # good old Redo!
                    self.console.write_line(b'?Redo from start')
                    readvar = var
                else:
                    varlist = [r + [v] for r, v in zip(var, values)]
                    break
            self.parser.redo_on_break = False
            self.interpreter.input_mode = False
            for v in varlist:
                self.memory.set_variable(*v)

    def _input_file(self, finp, readvar):
        """INPUT: retrieve input from file."""
        for v in readvar:
            name, indices = v
            typechar = self.memory.complete_name(name)[-1:]
            word, _ = finp.input_entry(typechar, allow_past_end=False)
            value = self.values.from_repr(word, allow_nonnum=True, typechar=typechar)
            self.memory.set_variable(name, indices, value)

    def line_input_(self, args):
        """LINE INPUT: request line of input from user."""
        file_number = next(args)
        if file_number is None:
            # get prompt
            newline, prompt, _ = next(args)
            finp = None
        else:
            prompt, newline = None, None
            file_number = values.to_int(file_number)
            error.range_check(0, 255, file_number)
            finp = self.files.get(file_number, mode=b'IR')
        # get string variable
        readvar, indices = next(args)
        list(args)
        readvar = self.memory.complete_name(readvar)
        if readvar[-1:] != values.STR:
            raise error.BASICError(error.TYPE_MISMATCH)
        # read the input
        if finp:
            line, cr = finp.read_line()
            if not line and not cr:
                raise error.BASICError(error.INPUT_PAST_END)
        else:
            self.interpreter.input_mode = True
            self.parser.redo_on_break = True
            line = self.console.read_line(prompt, write_endl=newline, is_input=True)
            self.parser.redo_on_break = False
            self.interpreter.input_mode = False
        self.memory.set_variable(readvar, indices, self.values.from_value(line, values.STR))

    def randomize_(self, args):
        """RANDOMIZE: set random number generator seed."""
        val, = args
        if val is not None:
            # don't convert to int if provided in the code
            val = values.pass_number(val, err=error.IFC)
        else:
            # prompt for random seed if not specified
            while True:
                seed = self.console.read_line(
                    b'Random number seed (-32768 to 32767)? ', is_input=True
                )
                try:
                    val = self.values.from_repr(seed, allow_nonnum=False)
                except error.BASICError as e:
                    if e.err != error.IFC:
                        raise
                else:
                    break
            # seed entered on prompt is rounded to int
            val = values.to_integer(val)
        self.randomiser.reseed(val)

    def key_(self, args):
        """KEY: macro or event trigger definition."""
        keynum = values.to_int(next(args))
        error.range_check(1, 255, keynum)
        text = values.next_string(args)
        list(args)
        try:
            self.console.set_macro(keynum, text)
            return
        except ValueError:
            pass
        # if out of range of number of macros (12 on Tandy, else 10), it's a trigger definition
        try:
            self.basic_events.key[keynum-1].set_trigger(text)
        except IndexError:
            # out of range key value
            # if the text is  two letters long (as for a trigger definition), no error is raised
            if len(text) != 2:
                raise error.BASICError(error.IFC)

    def pen_fn_(self, args):
        """PEN: poll the light pen."""
        fn, = args
        result = self.pen.poll(fn, self.basic_events.pen in self.basic_events.enabled, self.display.apage)
        return self.values.new_integer().from_int(result)



"""
PC-BASIC - tokeniser.py
Convert plain-text BASIC code to tokenised form

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import struct
import io

from ..base import tokens as tk
from ..base.tokens import DIGITS, LETTERS
from ..base import codestream
from .. import values


class PlainTextStream(codestream.CodeStream):
    """Stream of plain-text BASIC code."""

    end_line = (b'\0', b'\r')

    def read_line_number(self):
        """Read a line or jump number, return as int."""
        word = bytearray()
        ndigits, nblanks = 0, 0
        # don't read more than 5 digits
        while (ndigits < 5):
            c = self.peek()
            if not c:
                break
            elif c in DIGITS:
                word += self.read(1)
                nblanks = 0
                ndigits += 1
                if int(word) > 6552:
                    # note: anything >= 65530 is illegal in GW-BASIC
                    # in loading an ASCII file, GWBASIC would interpret these as
                    # '6553 1' etcetera, generating a syntax error on load.
                    break
            elif c in self.blanks:
                self.read(1)
                nblanks += 1
            else:
                break
        # don't claim trailing w/s
        self.seek(-nblanks, 1)
        if word:
            return int(word)
        return None


class Tokeniser(object):
    """BASIC tokeniser."""

    # keywords than can followed by one or more line numbers
    _linenum_words = (
        tk.KW_GOTO, tk.KW_THEN, tk.KW_ELSE, tk.KW_GOSUB,
        tk.KW_LIST, tk.KW_RENUM, tk.KW_EDIT, tk.KW_LLIST,
        tk.KW_DELETE, tk.KW_RUN, tk.KW_RESUME, tk.KW_AUTO,
        tk.KW_ERL, tk.KW_RESTORE, tk.KW_RETURN)

    # operator symbols
    _ascii_operators = b'+-=/\\^*<>'

    def __init__(self, values, keyword_dict):
        """Initialise tokeniser."""
        self._values = values
        self._keyword_to_token = keyword_dict.to_token

    def tokenise_line(self, line):
        """Convert an ascii program line to tokenised form."""
        ins = PlainTextStream(line)
        outs = codestream.TokenisedStream()
        # skip whitespace at start of line
        d = ins.skip_blank()
        if d == b'':
            # empty line at EOF
            return outs
        # read the line number
        self._tokenise_line_number(ins, outs)
        # expect line number
        allow_jumpnum = False
        # expect number (6553 6 -> the 6 is encoded as \x17)
        allow_number = True
        # flag for SPC( or TAB( as numbers can follow the closing bracket
        spc_or_tab = False
        # parse through elements of line
        while True:
            # peek next character
            c = ins.peek()
            # anything after NUL is ignored till EOL
            if c == b'\0':
                ins.read(1)
                ins.read_to((b'', b'\r'))
                break
            # end of line
            elif c in (b'', b'\r'):
                break
            # handle whitespace
            elif c in ins.blanks:
                ins.read(1)
                outs.write(c)
            # handle string literals
            elif ins.peek() == b'"':
                outs.write(ins.read_string())
            # handle jump numbers
            elif allow_number and allow_jumpnum and c in DIGITS + b'.':
                self._tokenise_jump_number(ins, outs)
            # handle numbers
            # numbers following var names with no operator or token in between
            # should not be parsed, eg OPTION BASE 1
            # note we don't include leading signs, encoded as unary operators
            # number starting with & are always parsed
            elif c in (b'&', ) or (
                    allow_number and not allow_jumpnum and c in DIGITS + b'.'
                ):
                outs.write(self._tokenise_number(ins))
            # operator keywords ('+', '-', '=', '/', '\\', '^', '*', '<', '>'):
            elif c in self._ascii_operators:
                ins.read(1)
                # operators don't affect line number mode - can do line number
                # arithmetic and RENUM will do the strangest things
                # this allows for 'LIST 100-200' etc.
                outs.write(self._keyword_to_token[c])
                allow_number = True
            # special case ' -> :REM'
            elif c == b"'":
                ins.read(1)
                outs.write(b':' + tk.REM + tk.O_REM)
                self._tokenise_rem(ins, outs)
            # special case ? -> PRINT
            elif c == b'?':
                ins.read(1)
                outs.write(tk.PRINT)
                allow_number = True
            # keywords & variable names
            elif c in LETTERS:
                word = self._tokenise_word(ins, outs)
                # handle non-parsing modes
                if word in (tk.KW_REM, b"'"):
                    self._tokenise_rem(ins, outs)
                elif word == tk.KW_DATA:
                    self._tokenise_data(ins, outs)
                else:
                    allow_jumpnum = (word in self._linenum_words)
                    # numbers can follow tokenised keywords
                    # (which does not include the word 'AS')
                    allow_number = (word in self._keyword_to_token)
                    if word in (tk.KW_SPC, tk.KW_TAB):
                        spc_or_tab = True
            else:
                ins.read(1)
                if c in (b',', b'#', b';'):
                    # can separate numbers as well as jumpnums
                    allow_number = True
                elif c in (b'(', b'['):
                    allow_number = True
                elif c == b')':
                    if spc_or_tab:
                        spc_or_tab = False
                        allow_jumpnum = False
                    # closing bracket does not affect line number mode
                    allow_number = True
                else:
                    allow_jumpnum, allow_number = False, False
                # replace all other nonprinting chars by spaces;
                # HOUSE 0x7f is allowed.
                outs.write(c if ord(c) >= 32 and ord(c) <= 127 else b' ')
        outs.seek(0)
        return outs

    def _tokenise_rem(self, ins, outs):
        """Pass anything after REM as is till EOL."""
        outs.write(ins.read_to((b'', b'\r', b'\0')))

    def _tokenise_data(self, ins, outs):
        """Pass DATA as is, till end of statement, except for literals."""
        while True:
            outs.write(ins.read_to((b'', b'\r', b'\0', b':', b'"')))
            if ins.peek() == b'"':
                # string literal in DATA
                outs.write(ins.read_string())
            else:
                break

    def _tokenise_line_number(self, ins, outs):
        """Convert an ascii line number to tokenised start-of-line."""
        linenum = ins.read_line_number()
        if linenum is not None:
            # NUL terminates last line and fills up the first char in the buffer
            # (that would be the magic number when written to file)
            # in direct mode, we'll know to expect a line number if the output
            # starts with a NUL
            # next two bytes are for internal use and at this point
            # can be anything nonzero; we use this.
            outs.write(b'\x00\xC0\xDE' + struct.pack('<H', linenum))
            # ignore single whitespace after line number, if any,
            # unless line number is zero (as does GW)
            if ins.peek() == b' ' and linenum != 0:
                ins.read(1)
        else:
            # direct line; internally, we need an anchor for the program pointer,
            # so we encode a ':'
            outs.write(b':')

    def _tokenise_jump_number(self, ins, outs):
        """Convert an ascii line number pointer to tokenised form."""
        linum = ins.read_line_number()
        if linum is not None:
            outs.write(tk.T_UINT + struct.pack('<H', linum))
        elif ins.peek() == b'.':
            ins.read(1)
            outs.write(b'.')

    def _tokenise_word(self, ins, outs):
        """Convert a keyword or name to tokenised form."""
        word = b''
        while True:
            c = ins.read(1)
            word += c.upper()
            if word == b'GO':
                # deal with special cases 'GO     TO' -> 'GOTO', 'GO SUB' -> 'GOSUB'
                word, allow_name_chars = self._tokenise_wide_goto_gosub(ins)
            else:
                allow_name_chars = False
            if word in self._keyword_to_token:
                # ignore if part of a longer name, except FN, SPC(, TAB(, USR, GO SUB and GO   TO
                if word not in (tk.KW_FN, tk.KW_SPC, tk.KW_TAB, tk.KW_USR) and not allow_name_chars:
                    nxt = ins.peek()
                    if nxt and nxt in tk.NAME_CHARS:
                        continue
                token = self._keyword_to_token[word]
                # handle special case ELSE -> :ELSE
                if word == tk.KW_ELSE:
                    outs.write(b':' + token)
                # handle special case WHILE -> WHILE+
                elif word == tk.KW_WHILE:
                    outs.write(token + tk.O_PLUS)
                else:
                    outs.write(token)
                break
            # allowed names: letter + (letters, numbers, .)
            elif not c:
                outs.write(word)
                break
            elif c not in tk.NAME_CHARS:
                word = word[:-1]
                ins.seek(-1, 1)
                outs.write(word)
                break
        return word

    def _tokenise_wide_goto_gosub(self, ins):
        """Special cases 'GO     TO' -> 'GOTO', 'GO SUB' -> 'GOSUB'."""
        word = b'GO'
        allow_name_chars = False
        next_four = ins.peek(4).upper()
        # GO SUB allows 1 space, allows text after
        if next_four == b' SUB':
            word = tk.KW_GOSUB
            ins.read(4)
            allow_name_chars = True
        # GO TO with single space, does not allow text or numbers after
        elif next_four[:3] == b' TO' and next_four[3:4] not in tk.NAME_CHARS:
            word = tk.KW_GOTO
            ins.read(3)
        # GO  TO allows more than 1 spaces, but not \t or \n
        # and *then* allows text after
        elif next_four[:2] == b'  ':
            pos = ins.tell()
            while ins.peek(1) == b' ':
                ins.read(1)
            next_two = ins.read(2).upper()
            if next_two == b'TO':
                word = tk.KW_GOTO
                allow_name_chars = True
            else:
                ins.seek(pos)
        return word, allow_name_chars

    def _tokenise_number(self, ins):
        """Convert Python-string number representation to number token."""
        word = ins.read_number()
        if word[:2] == b'&H':
            # hex constant
            return self._values.new_integer().from_hex(word[2:]).to_token_hex()
        elif word[:2] == b'&O':
            # octal constant
            # read_number converts &1 into &O1
            return self._values.new_integer().from_oct(word[2:]).to_token_oct()
        elif word[:1] in DIGITS + b'.+-':
            # handle other numbers
            # note GW passes signs separately as a token
            # and only stores positive numbers in the program
            return self._values.from_repr(word, allow_nonnum=False).to_token()



"""
PC-BASIC - display.cursor
Cursor operations

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from ..base import signals


# text mode cursor:
#   cursor visibility is not affected by visible/active page
#   but if visible page is not active page, the visible cursor position is not updated (DOSBox)
#
#   in direct mode, the cursor is always visible
#
#   when a program is running:
#       the cursor is by default invisible
#       if LOCATE ,,1 is given the cursor is visible
#       on INPUT or SHELL, the cursor is visible
#
# graphics mode cursor:
#   the cursor is only visible if the active page is visible
#
#   in direct mode, the cursor is always visible
#
#   when a program is running:
#       the cursor is by default invisible
#       on INPUT or SHELL, the cursor is visible

# summary:
# active == visible or text mode:
#   - in direct mode, INPUT or SHELL the cursor is visible.
#   - if LOCATE,,1 and text mode, the cursor is visible.
# otherwise, the cursor is invisible



class Cursor(object):
    """Manage the cursor."""

    def __init__(self, queues, mode):
        """Initialise the cursor."""
        self._queues = queues
        self._mode = mode
        self._colourmap = None
        # are we in parse mode? invisible unless override_visible is True
        self._default_visible = True
        # cursor visible in parse mode? user override
        self._override_visible = False
        # actually visible at present
        # set to None to force a signal when first set to True or False
        self._visible = None
        # visibility override flags
        self._active = True
        self._direct = False
        self._override = False
        self._textmode_override = False
        # cursor shape
        self._from_line = 0
        self._to_line = 0
        self._width = self._mode.font_width
        self._height = self._mode.font_height
        self._fore_attr = None
        self._row, self._col = 1, 1

    def init_mode(self, mode, attr, colourmap):
        """Change the cursor for a new screen mode."""
        self._mode = mode
        self._height = mode.font_height
        self._colourmap = colourmap
        self._visible = None
        # set the cursor position and attribute
        self.move(1, 1, attr, new_width=1)
        # cursor width starts out as single char
        self.set_default_shape(True)
        self._set_visibility()

    def rebuild(self):
        """Rebuild the cursor on resume."""
        self._queues.video.put(signals.Event(
            signals.VIDEO_SET_CURSOR_SHAPE, (self._from_line, self._to_line)
        ))
        self._queues.video.put(signals.Event(
            signals.VIDEO_MOVE_CURSOR, (self._row, self._col, self._fore_attr, self._width)
        ))
        # set visibility and blink state
        # cursor blinks if and only if in text mode
        self._queues.video.put(signals.Event(
            signals.VIDEO_SHOW_CURSOR, (self._visible, self._mode.is_text_mode)
        ))

    # attribute

    def set_attr(self, new_attr):
        """Set the text cursor attribute and submit."""
        self.move(self._row, self._col, new_attr, None)

    # location

    def move(self, new_row, new_column, new_attr=None, new_width=None):
        """Move the cursor and submit."""
        if new_attr:
            fore, _, _, _ = self._colourmap.split_attr(new_attr)
        else:
            fore = self._fore_attr
        if not new_width:
            new_width = self._width
        else:
            new_width = new_width * self._mode.font_width
        if (
                (new_row, new_column) == (self._row, self._col)
                and fore == self._fore_attr and new_width == self._width
            ):
            return
        # only submit move signal if visible (so that we see it in the right place)
        # or if the row changes (so that row-based cli interface can keep up with current row
        if self._visible or new_row != self._row:
            self._queues.video.put(signals.Event(
                signals.VIDEO_MOVE_CURSOR, (new_row, new_column, fore, new_width)
            ))
        self._row, self._col = new_row, new_column
        self._fore_attr = fore
        self._width = new_width

    # visibility

    def set_active(self, active):
        """Active page is visible page, so we have a cursor."""
        self._active = active
        self._set_visibility()

    def set_direct(self, direct):
        """Direct mode - so cursor is visible."""
        self._direct = direct
        self._set_visibility()

    def set_override(self, override):
        """INPUT - so cursor is visible."""
        self._override = override
        self._set_visibility()

    def set_textmode_override(self, override):
        """LOCATE,,1 - so cursor is visible."""
        self._textmode_override = override
        self._set_visibility()

    def _set_visibility(self):
        """Set cursor visibility to its default state."""
        visible = self._active and (
            self._direct or self._override or (
                self._textmode_override and self._mode.is_text_mode
            )
        )
        if self._visible != visible:
            self._visible = visible
            if visible:
                # update position, attribute and shape
                self._queues.video.put(signals.Event(
                    signals.VIDEO_MOVE_CURSOR, (self._row, self._col, self._fore_attr, self._width)
                ))
            # show or hide the cursor and set blink
            self._queues.video.put(signals.Event(
                signals.VIDEO_SHOW_CURSOR, (visible, self._mode.is_text_mode)
            ))

    # shape

    @property
    def shape(self):
        """Cursor shape (from, to)."""
        return self._from_line, self._to_line

    def set_shape(self, from_line, to_line):
        """Set the cursor shape."""
        # odd treatment of cursors on EGA/VGA machines (14-pixel and up),
        # presumably for backward compatibility
        # the following algorithm is based on DOSBox source int10_char.cpp
        #     INT10_SetCursorShape(Bit8u first,Bit8u last)
        if self._height > 9:
            max_line = self._height - 1
            if from_line & 0xe0 == 0 and to_line & 0xe0 == 0:
                if (to_line < from_line):
                    # invisible only if to_line is zero and to_line < from_line
                    if to_line != 0:
                        # block shape from *to_line* to end
                        from_line = to_line
                        to_line = max_line
                elif ((from_line | to_line) >= max_line or
                            to_line != max_line-1 or from_line != max_line):
                    if to_line > 3:
                        if from_line+2 < to_line:
                            if from_line > 2:
                                from_line = (max_line+1) // 2
                            to_line = max_line
                        else:
                            from_line = from_line - to_line + max_line
                            to_line = max_line
                            if max_line > 0xc:
                                from_line -= 1
                                to_line -= 1
        self._from_line = max(0, min(from_line, self._height-1))
        self._to_line = max(0, min(to_line, self._height-1))
        self._queues.video.put(signals.Event(
            signals.VIDEO_SET_CURSOR_SHAPE, (self._from_line, self._to_line))
        )

    def set_default_shape(self, overwrite_shape):
        """Set the cursor to one of two default shapes."""
        if overwrite_shape:
            # most modes have cursor on last line
            self.set_shape(*self._mode.cursor_shape)
        else:
            # half-block cursor for insert
            self.set_shape(self._height//2, self._height-1)



"""
PC-BASIC - compat.console
Console and standard I/O setup

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import atexit


from .base import WIN32


if WIN32:
    from .win32_console import Win32Console, IS_CONSOLE_APP
    from .win32_console import StdIO, init_stdio

    if IS_CONSOLE_APP:
        console = Win32Console()
    else:
        console = None

    stdio = StdIO(console)


    ##############################################################################
    # non-blocking input

    def read_all_available(stream):
        """Read all available bytes or unicode from a stream; nonblocking; None if closed."""
        # are we're reading from (wrapped) stdin or not?
        if hasattr(stream, 'isatty') and stream.isatty():
            # this is shaky - try to identify unicode vs bytes stream
            is_unicode_stream = hasattr(stream, 'buffer')
            # console always produces unicode
            unistr = console.read_all_chars()
            # but convert to bytes if the tty stream provides was a bytes stream
            if is_unicode_stream or unistr is None:
                return unistr
            else:
                return unistr.encode(stdio.stdin.encoding, 'replace')
        else:
            # this would work on unix too
            # just read the whole file and be done with it
            # bytes or unicode, depends on stream
            return stream.read() or None

else:
    from .posix_console import PosixConsole, read_all_available, IS_CONSOLE_APP
    from .posix_console import StdIO, init_stdio

    stdio = StdIO()

    try:
        console = PosixConsole(stdio)
    except EnvironmentError:
        console = None


    # don't crash into raw terminal
    def _atexit_unset_raw():
        try:
            console.unset_raw()
        except Exception:
            pass

    atexit.register(_atexit_unset_raw)



"""
PC-BASIC - numbers.py
Integer and Floating Point values

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3.
"""

# descriptions of the Microsoft Binary Format found here:
# http://www.experts-exchange.com/Programming/Languages/Pascal/Delphi/Q_20245266.html
# http://www.boyet.com/Articles/MBFSinglePrecision.html
#
# single precision:                      m3 | m2 | m1 | exponent
# double precision:  m7 | m6 | m5 | m4 | m3 | m2 | m1 | exponent
# where:
#     m1 is most significant byte => sbbb|bbbb
#     m7 is the least significant byte
#     m = mantissa byte
#     s = sign bit
#     b = bit
#
# The exponent is biased by 128.
# There is an assumed 1 bit after the radix point (so the assumed mantissa is 0.1ffff... where f's are the fraction bits)

import binascii
import struct
import math

from ...compat import iterchar, int2byte

from ..base import tokens as tk
from ..base import error


# for to_str
# for numbers, tab and LF are whitespace
BLANKS = b' \t\n'
# ASCII separators - these cause string representations to evaluate to zero
SEPARATORS = b'\x1c\x1d\x1f'



##############################################################################
# value base class

class Value(object):
    """Abstract base class for value types."""

    sigil = None
    size = None

    def __init__(self, buffer, values):
        """Initialise the value."""
        if buffer is None:
            buffer = bytearray(self.size)
        self._buffer = memoryview(buffer)
        self._values = values

    def __repr__(self):
        """String representation for debugging."""
        sigil_repr = self.sigil.decode('ascii')
        bytes_repr = binascii.hexlify(self.to_bytes()).decode('ascii')
        try:
            return '%s[%s %r]' % (
                sigil_repr, bytes_repr,
                self.to_value()
            )
        except Exception:
            return '%s[%s <detached>]' % (sigil_repr, bytes_repr)

    def __getstate__(self):
        # can't pickle memoryview
        self._buffer = bytearray(self._buffer)
        return self.__dict__

    def __setstate__(self, pickle_dict):
        # can't pickle memoryview
        self.__dict__ = pickle_dict
        self._buffer = memoryview(self._buffer)

    def to_value(self):
        """Convert to Python value."""

    def clone(self):
        """Create a copy."""
        return self.__class__(None, self._values).from_bytes(self._buffer)

    def new(self):
        """Create a new null value."""
        return self.__class__(None, self._values)

    def copy_from(self, other):
        """Copy another value into this one."""
        self._buffer[:] = other._buffer
        return self

    def to_bytes(self):
        """Get a copy of the byte representation."""
        return bytearray(self._buffer)

    def from_bytes(self, in_bytes):
        """Copy a new byte representation into the value."""
        self._buffer[:] = in_bytes
        return self

    def view(self):
        """Get a reference to the storage space."""
        return self._buffer


##############################################################################
# numeric value base class

class Number(Value):
    """Abstract base class for numeric value."""

    zero = None
    pos_max = None
    neg_max = None

    def __init__(self, buffer, values):
        """Initialise the number."""
        Value.__init__(self, buffer, values)
        self.error_handler = values.error_handler

    def to_double(self):
        """Convert to double."""

    def to_single(self):
        """Convert to single."""

    def to_float(self, allow_double=True):
        """Convert to float."""

    def add(self, rhs):
        """Add another Number."""
        return self.clone().iadd(rhs)

    def iadd(self, rhs):
        """Add another Number, in-place."""


##############################################################################
# integer number

class Integer(Number):
    """16-bit signed little-endian integer."""

    sigil = b'%'
    size = 2

    pos_max = b'\xff\x7f'
    neg_max = b'\xff\xff'

    def is_zero(self):
        """Value is zero."""
        return self._buffer == b'\0\0'

    def is_negative(self):
        """Value is negative."""
        return (bytearray(self._buffer)[-1] & 0x80) != 0

    def sign(self):
        """Sign of value."""
        if (bytearray(self._buffer)[-1] & 0x80) != 0:
            return -1
        elif bytearray(self._buffer) == bytearray(b'\0\0'):
            return 0
        else:
            return 1

    def to_int(self, unsigned=False):
        """Return value as Python int."""
        if unsigned:
            return struct.unpack('<H', self._buffer)[0]
        else:
            return struct.unpack('<h', self._buffer)[0]

    def from_int(self, in_int, unsigned=False):
        """Set value to Python int."""
        if unsigned:
            # we can in fact assign negatives as 'unsigned'
            if in_int < 0:
                in_int += 0x10000
            intformat = '<H'
            maxint = 0xffff
        else:
            intformat = '<h'
            maxint = 0x7fff
        if not (-0x8000 <= in_int <= maxint):
            raise error.BASICError(error.OVERFLOW)
        struct.pack_into(intformat, self._buffer, 0, in_int)
        return self

    def to_integer(self, unsigned=False):
        """Convert to Integer (no-op)."""
        return self

    def to_double(self):
        """Convert to double."""
        return Double(None, self._values).from_integer(self)

    def to_single(self):
        """Convert to single."""
        return Single(None, self._values).from_integer(self)

    def to_float(self, allow_double=True):
        """Convert to float."""
        return Single(None, self._values).from_integer(self)

    to_value = to_int
    from_value = from_int

    def to_token(self):
        """Return signed value as integer token."""
        if bytearray(self._buffer)[1] == 0:
            byte = bytearray(self._buffer)[0]
            # although there is a one-byte token for '10', we don't write it.
            if byte < 10:
                return int2byte(ord(tk.C_0) + byte)
            else:
                return tk.T_BYTE + int2byte(bytearray(self._buffer)[0])
        else:
            return tk.T_INT + self._buffer.tobytes()

    def to_token_hex(self):
        """Return unsigned value as hex token."""
        return tk.T_HEX + self._buffer.tobytes()

    def to_token_oct(self):
        """Return unsigned value as oct token."""
        return tk.T_OCT + self._buffer.tobytes()

    def from_token(self, token):
        """Set value to signed or unsigned integer token."""
        d = bytearray(token)[0]
        if d in (ord(_c) for _c in (tk.T_OCT, tk.T_HEX, tk.T_INT, tk.T_UINT)):
            self._buffer[:] = token[-2:]
        elif d == ord(tk.T_BYTE):
            self._buffer[:] = token[-1:] + b'\0'
        elif ord(tk.C_0) <= d <= ord(tk.C_10):
            self._buffer[:] = int2byte(d - 0x11) + b'\0'
        else:
            raise ValueError('%s is not an Integer token.' % repr(token))
        return self

    # representations

    def to_oct(self):
        """Convert integer to str in octal representation."""
        if self.is_zero():
            return b'0'
        return b'%o' % (self.to_int(unsigned=True),)

    def to_hex(self):
        """Convert integer to str in hex representation."""
        return b'%X' % (self.to_int(unsigned=True),)

    def to_str(self, leading_space, type_sign):
        """Convert integer to str in decimal representation."""
        intstr = b'%d' % self.to_int()
        if leading_space and intstr[0:1] != b'-':
            return b' ' + intstr
        else:
            return intstr

    def from_oct(self, oct_repr):
        """Convert str in octal representation to integer."""
        # oct representations may be interrupted by blanks
        val = int(oct_repr.strip(BLANKS), 8) if oct_repr.strip(BLANKS) else 0
        return self.from_int(val, unsigned=True)

    def from_hex(self, hex_repr):
        """Convert str in hexadecimal representation to integer."""
        # hex representations must be contiguous
        val = int(hex_repr, 16) if hex_repr else 0
        return self.from_int(val, unsigned=True)

    def from_str(self, dec_repr):
        """Convert str in decimal representation to integer."""
        valstr = dec_repr.strip(BLANKS)
        if set(valstr) - set(tk.DIGITS):
            raise ValueError(
                'string %r contains non-digits, cannot convert to Integer.' %
                (valstr,)
            )
        return self.from_int(int(valstr))

    # operations

    def itrunc(self):
        """Truncate towards zero in-place (no-op)."""
        return self

    def ifloor(self):
        """Truncate towards negative infinity in-place (no-op)."""
        return self

    def ineg(self):
        """Negate in-place."""
        if self._buffer == b'\x00\x80':
            raise error.BASICError(error.OVERFLOW)
        lsb = (bytearray(self._buffer)[0] ^ 0xff) + 1
        msb = bytearray(self._buffer)[1] ^ 0xff
        # apply carry
        if lsb > 0xff:
            lsb -= 0x100
            msb += 1
        # ignore overflow, since -0 == 0
        self._buffer[:] = bytearray([lsb, msb & 0xff])
        return self

    def iabs(self):
        """Absolute value in-place."""
        if bytearray(self._buffer)[-1] & 0x80:
            return self.ineg()
        return self

    def iadd(self, rhs):
        """Add another Integer in-place."""
        lsb = bytearray(self._buffer)[0] + bytearray(rhs._buffer)[0]
        msb = bytearray(self._buffer)[1] + bytearray(rhs._buffer)[1]
        # apply carry
        if lsb > 0xff:
            lsb -= 0x100
            msb += 1
        # overflow if signs were equal and have changed
        if (
                (bytearray(self._buffer)[1] > 0x7f) ==
                (bytearray(rhs._buffer)[1] > 0x7f) !=
                (msb > 0x7f)
            ):
            raise error.BASICError(error.OVERFLOW)
        self._buffer[:] = bytearray([lsb, msb & 0xff])
        return self

    def isub(self, rhs):
        """Subtract another Integer in-place."""
        # we can't apply the neg on the lhs and avoid a copy
        # because of things like -32768 - (-1)
        return self.iadd(rhs.clone().ineg())

    # no imul - we always promote to float first for multiplication
    # no idiv - we always promote to float first for true division

    def idiv_int(self, rhs):
        """Perform integer division."""
        if rhs.is_zero():
            # division by zero - return single-precision maximum
            if self.is_negative():
                max_val = Single(None, self._values).from_bytes(Single.neg_max)
            else:
                max_val = Single(None, self._values).from_bytes(Single.pos_max)
            raise ZeroDivisionError(max_val)
        dividend = self.to_int()
        divisor = rhs.to_int()
        # BASIC intdiv rounds to zero, Python's floordiv to -inf
        if (dividend >= 0) == (divisor >= 0):
            return self.from_int(dividend // divisor)
        else:
            return self.from_int(-(abs(dividend) // abs(divisor)))

    def imod(self, rhs):
        """Left modulo right."""
        if rhs.is_zero():
            # division by zero - return single-precision maximum
            if self.is_negative():
                max_val = Single(None, self._values).from_bytes(Single.neg_max)
            else:
                max_val = Single(None, self._values).from_bytes(Single.pos_max)
            raise ZeroDivisionError(max_val)
        dividend = self.to_int()
        divisor = rhs.to_int()
        # BASIC MOD has same sign as dividend, Python mod has same sign as divisor
        mod = dividend % divisor
        if dividend < 0 or mod < 0:
            mod -= divisor
        return self.from_int(mod)

    # relations

    def gt(self, rhs):
        """Greater than."""
        if isinstance(rhs, Float):
            # upgrade to Float
            return rhs.new().from_integer(self).gt(rhs)
        isneg = bytearray(self._buffer)[-1] & 0x80
        if isneg != (bytearray(rhs._buffer)[-1] & 0x80):
            return not(isneg)
        # compute the unsigned (not absolute!) >
        lmsb = bytearray(self._buffer)[1] & 0x7f
        rmsb = bytearray(rhs._buffer)[1] & 0x7f
        if lmsb > rmsb:
            return True
        elif lmsb < rmsb:
            return False
        return bytearray(self._buffer)[0] > bytearray(rhs._buffer)[0]

    def eq(self, rhs):
        """Equals."""
        if isinstance(rhs, Float):
            # upgrade to Float
            return rhs.new().from_integer(self).eq(rhs)
        return self._buffer == rhs._buffer


##############################################################################
# floating-point base class

class Float(Number):
    """Abstract base class for floating-point value."""

    digits = None
    pos_max = None
    neg_max = None

    exp_sign = None

    # properties

    def is_zero(self):
        """Value is zero."""
        return bytearray(self._buffer)[-1] == 0

    def is_negative(self):
        """Value is negative."""
        return bytearray(self._buffer)[-2] >= 0x80

    def sign(self):
        """Sign of value."""
        if bytearray(self._buffer)[-1] == 0:
            return 0
        elif (bytearray(self._buffer)[-2] & 0x80) != 0:
            return -1
        return 1

    # BASIC type conversions

    def from_integer(self, in_integer):
        """Convert Integer to Float."""
        return self.from_int(in_integer.to_int())

    def to_integer(self, unsigned=False):
        """Convert Float to Integer."""
        return Integer(None, self._values).from_int(self.to_int(), unsigned)

    # Python float conversions

    def to_value(self):
        """Return value as Python float."""
        exp = bytearray(self._buffer)[-1] - self._bias
        if exp == -self._bias:  # pylint: disable=invalid-unary-operand-type
            return 0.
        # unpack as unsigned long int
        man = struct.unpack(self._intformat, bytearray(self._buffer)[:-1] + b'\0')[0]
        # prepend assumed bit and apply sign
        if man & self._signmask:
            man = -man
        else:
            man |= self._signmask
        return man * 2.**exp

    def from_value(self, in_float):
        """Set to value of Python float."""
        if in_float == 0.:
            self._buffer[:] = b'\0' * self.size
            return self
        neg = in_float < 0
        exp = int(math.log(abs(in_float), 2) - self._shift)
        man = int(abs(in_float) * 0.5**exp)
        exp += self._bias
        # why do we need this here?
        man, exp = self._bring_to_range(man, exp, self._posmask, self._mask)
        if not self._check_limits(exp, neg):
            return self
        struct.pack_into(
            self._intformat, self._buffer, 0, man & (self._mask if neg else self._posmask)
        )
        self._buffer[-1:] = int2byte(exp)
        return self


    # Python int conversions

    def to_int(self):
        """Return value rounded to Python int."""
        man, neg = self._to_int_den()
        # carry: round to nearest, halves away from zero
        if man & 0x80:
            man += 0x80
        # note that -man >> 8 and -(man >> 8) are different
        # due to Python's rounding rules
        # python2: int() converts from type long if possible
        return int(-(man >> 8) if neg else man >> 8)

    def from_int(self, in_int):
        """Set value to Python int."""
        if in_int == 0:
            self._buffer[:] = b'\0' * self.size
        else:
            neg = in_int < 0
            man, exp = self._bring_to_range(abs(in_int), self._bias, self._posmask, self._mask)
            if not self._check_limits(exp, neg):
                return self
            struct.pack_into(
                self._intformat, self._buffer, 0, man & (self._mask if neg else self._posmask)
            )
            self._buffer[-1:] = int2byte(exp)
        return self

    def to_int_truncate(self):
        """Truncate float to integer."""
        man, neg = self._to_int_den()
        # discard carry
        # python2: int() converts from type long if possible
        return int(-(man >> 8) if neg else man >> 8)

    def mantissa(self):
        """Integer value of mantissa."""
        exp, man, neg = self._denormalise()
        return -(man >> 8) if neg else (man >> 8)

    # in-place unary operations

    def ineg(self):
        """Negate in-place."""
        self._buffer[-2:-1] = int2byte(bytearray(self._buffer)[-2] ^ 0x80)
        return self

    def iabs(self):
        """Absolute value in-place."""
        self._buffer[-2:-1] = int2byte(bytearray(self._buffer)[-2] & 0x7F)
        return self

    def itrunc(self):
        """Truncate towards zero in-place."""
        return self.from_int(self.to_int_truncate())

    def ifloor(self):
        """Truncate towards negative infinity in-place."""
        oldval = self.clone()
        # we need to check negativity first, as itrunc() may take us to zero
        was_negative = self.is_negative()
        self.itrunc()
        if not self.eq(oldval) and was_negative:
            self.isub(self.new().from_bytes(self._one))
        return self

    # relations

    def gt(self, rhs):
        """Greater than."""
        if isinstance(rhs, Integer):
            # upgrade rhs to Float
            return self.gt(self.new().from_integer(rhs))
        elif isinstance(rhs, Double) and isinstance(self, Single):
            # upgrade to Double
            return Double(None, self._values).from_single(self).gt(rhs)
        rhsneg = rhs.is_negative()
        # treat zero separately to avoid comparing different mantissas
        # zero is only greater than negative
        if self.is_zero():
            return bool(rhsneg) and not rhs.is_zero()
        isneg = self.is_negative()
        if isneg != rhsneg:
            return not(isneg)
        if isneg:
            return rhs._abs_gt(self)
        return self._abs_gt(rhs)

    def eq(self, rhs):
        """Equals."""
        if isinstance(rhs, Integer):
            # upgrade rhs to Float
            return self.eq(self.new().from_integer(rhs))
        elif isinstance(rhs, Double) and isinstance(self, Single):
            # upgrade to Double
            return Double(None, self._values).from_single(self).eq(rhs)
        # all zeroes are equal
        if self.is_zero():
            return rhs.is_zero()
        return self._buffer == rhs._buffer

    # in-place binary operations

    def iadd(self, right):
        """Add in-place."""
        return self._normalise(*self._add_den(self._denormalise(), right._denormalise()))

    def isub(self, right):
        """Subtract in-place."""
        rexp, rman, rneg = right._denormalise()
        return self._normalise(*self._add_den(self._denormalise(), (rexp, rman, not rneg)))

    def imul(self, right_in):
        """Multiply in-place."""
        if self.is_zero() or right_in.is_zero():
            # set any zeroes to standard zero
            self._buffer[:] = b'\0' * self.size
            return self
        lexp, lman, lneg = self._denormalise()
        rexp, rman, rneg = right_in._denormalise()
        lexp += rexp - right_in._bias - 8
        lneg = (lneg != rneg)
        lman *= rman
        if lexp < -31:
            self._buffer[:] = b'\0' * self.size
            return self
        # drop some precision
        lman, lexp = self._bring_to_range(lman, lexp, self._den_mask>>4, self._den_upper>>4)
        # rounding quirk
        if lman & 0xf == 0x9:
            lman &= (self._carrymask + 0xfe)
        self._normalise(lexp, lman, lneg)
        return self

    def idiv(self, right_in):
        """Divide in-place."""
        if right_in.is_zero():
            # division by zero - return max float with the type and sign of self
            self.from_bytes(self.neg_max if self.is_negative() else self.pos_max)
            raise ZeroDivisionError(self)
        if self.is_zero():
            return self
        lexp, lman, lneg = self._div_den(self._denormalise(), right_in._denormalise())
        # normalise and return
        return self._normalise(lexp, lman, lneg)

    def ipow_int(self, expt):
        """Raise to integer power in-place."""
        return self._ipow_int(expt.to_int())


    # decimal representations

    def to_str(self, leading_space, type_sign):
        """Convert to string representation."""
        if self.is_zero():
            return (b' ' * leading_space) + b'0' + (self.sigil * type_sign)
        # sign or leading space
        sign = b''
        if self.is_negative():
            sign = b'-'
        elif leading_space:
            sign = b' '
        mantissa, exp10 = self.to_decimal(self.digits)
        digitstr = _get_digits(mantissa, self.digits).rstrip(b'0')
        # exponent for scientific notation
        exp10 += self.digits - 1
        if exp10 > self.digits - 1 or len(digitstr)-exp10 > self.digits + 1:
            # use scientific notation
            # this is self.to_str_scientific(
            #     n_before=1, n_decimals=self.digits-1, force_dot=False)
            valstr = self._scientific_notation(digitstr, exp10, digits_to_dot=1, force_dot=False)
        else:
            # use decimal notation
            # usual: n_decimals = -exp10
            # this is self.to_str_fixed
            #     (self, n_decimals=-exp10, force_dot=False, group_thousands=False) but with type_sign
            valstr = self._decimal_notation(digitstr, exp10, type_sign, force_dot=False)
        return sign + valstr

    def to_decimal(self, digits):
        """Return value as mantissa of length min(digits, self.digits) and decimal exponent."""
        if digits >= self.digits:
            # we'd be better off storing these (and self._ten) in denormalised form
            lim_bot = self.new().from_bytes(self._lim_bot)
            lim_top = self.new().from_bytes(self._lim_top)
        elif digits > 0:
            lim_bot = self.new().from_int(10**(digits-1))._just_under()
            lim_top = self.new().from_int(10**digits)._just_under()
        else:
            # use values for digits == 0 also if digits < 0
            lim_bot = self.new().from_int(0)
            lim_top = self.new().from_int(1)._just_under()
        tden = lim_top._denormalise()
        bden = lim_bot._denormalise()
        exp10 = 0
        den = self._denormalise()
        while self._abs_gt_den(den, tden):
            den = self._div10_den(den)
            exp10 += 1
        # rounding - gets us close to GW results
        den = self._apply_carry_den(den)
        while self._abs_gt_den(bden, den):
            den = self._mul10_den(den)
            exp10 -= 1
        # rounding - gets us close to GW results
        den = self._apply_carry_den(den)
        # round to int
        exp, man, neg = den
        exp -= self._bias
        if exp > 0: # pragma: no cover
            # I don't think this happens
            man <<= exp
        else:
            man >>= -exp
        # here we don't care about overflowing the mantissa as we output it as int
        if man & 0x80:
            man += 0x80
        num = -(man >> 8) if neg else man >> 8
        return num, exp10

    def from_decimal(self, mantissa, exp10):
        """Set value to mantissa and decimal exponent."""
        den = self.from_int(mantissa)._denormalise()
        # apply decimal exponent
        while (exp10 < 0):
            den = self._div10_den(den)
            exp10 += 1
        while (exp10 > 0):
            den = self._mul10_den(den)
            exp10 -= 1
        return self._normalise(*den)

    def to_str_scientific(self, digits_before_radix, digits_after_radix, always_show_radix):
        """Put a positive float in scientific format."""
        if self.is_zero():
            if always_show_radix:
                return b''.join((b'.', b'0' * digits_after_radix, self.exp_sign, b'+00'))
            # single/double difference: this matches GW output. odd, odd, odd
            if self.exp_sign == b'E':
                return b'E+00'
            return b'0D+00'
        digits_requested = digits_before_radix + digits_after_radix
        digits_precision = self.digits
        work_digits = min(digits_precision, digits_requested)
        # convert to mantissa of length work_digits and exponent
        mantissa, exponent = self.to_decimal(work_digits)
        # radix position in the digit string
        radix_position = exponent + work_digits
        digitstr = _get_digits(mantissa, work_digits)
        # ensure digitstr has the exact number of digits we're going to show
        # when work_digits == 0, e.g. PRINT USING "#^^^^";1 gives " E+01" not " E+00"
        # we get this because the digitstr is empty
        digitstr = digitstr.ljust(digits_requested, b'0')[:digits_requested]
        return self._scientific_notation(
            digitstr,
            # here, exp10 is the exponent if the radix is put after the first digit
            exp10=radix_position-1,
            digits_to_dot=digits_before_radix,
            force_dot=always_show_radix
        )

    def to_str_fixed(self, n_decimals, force_dot, group_thousands):
        """Put a float in fixed-point representation."""
        if self.is_zero():
            if force_dot:
                return b'.' + b'0'*n_decimals
            if n_decimals:
                return b'0'*n_decimals
            return b'0'
        # convert to integer_mantissa * 10**exponent
        mantissa, exp10 = self.to_decimal(self.digits)
        # -exp10 is the number of digits after the radix point
        n_after = -exp10
        # bring to decimal form of working precision
        if n_after > n_decimals:
            n_work = self.digits - (n_after - n_decimals)
            # this has n_work or n_work+1 digits, depending on rounding
            mantissa, exp10 = self.to_decimal(n_work)
            n_after = -exp10
        digitstr = b'%d' % (abs(mantissa),)
        # number of digits before the radix point.
        n_before = len(digitstr) - n_after
        # fill up with zeros to required number of figures
        digitstr = digitstr.ljust(n_decimals + n_before, b'0')
        return self._decimal_notation(
            digitstr, n_before-1, type_sign=b'',
            force_dot=force_dot, group_thousands=group_thousands
        )

    # implementation: floating- and fixed-point decimal notations

    def _group_thousands(self, digitstr):
        """Insert commas to group digits by thousands in a decimal number."""
        first = len(digitstr) % 3
        chunks = [digitstr[i:i + 3] for i in range(first, len(digitstr), 3)]
        if first:
            chunks = [digitstr[:first]] + chunks
        return b','.join(chunks)

    def _scientific_notation(self, digitstr, exp10, digits_to_dot, force_dot):
        """Put digits in scientific E-notation."""
        valstr = digitstr[:digits_to_dot]
        if len(digitstr) > digits_to_dot:
            after_str = digitstr[digits_to_dot:]
            valstr += b'.' + after_str
        elif len(digitstr) == digits_to_dot and force_dot:
            valstr += b'.'
        exponent = exp10 - digits_to_dot + 1
        valstr += self.exp_sign
        if exponent < 0:
            valstr += b'-'
        else:
            valstr += b'+'
        valstr += _get_digits(abs(exponent), min_digits=2)
        return valstr

    def _decimal_notation(self, digitstr, exp10, type_sign, force_dot, group_thousands=False):
        """Put digits in decimal notation."""
        type_sign = b'' if not type_sign else self.sigil
        # digits to decimal point
        exp10 += 1
        if exp10 >= len(digitstr):
            valstr = digitstr + b'0'*(exp10-len(digitstr))
            if group_thousands:
                valstr = self._group_thousands(valstr)
            if force_dot:
                valstr += b'.'
        elif exp10 > 0:
            valstr = digitstr[:exp10]
            if group_thousands:
                valstr = self._group_thousands(valstr)
            valstr += b'.' + digitstr[exp10:]
        else:
            valstr = b'.' + b'0'*(-exp10) + digitstr
        if (b'.' not in valstr) or (type_sign == b'#'):
            valstr += type_sign
        return valstr


    ##########################################################################
    # implementation: decimal representations only

    _one = None
    _ten = None
    _lim_bot = None
    _lim_top = None

    def _apply_carry_den(self, den):
        """Round the carry byte (to be used only in to_decimal)."""
        exp, man, neg = den
        # apply_carry()
        # carry bit set? then round up
        if (man & 0xff) > 0x7f:
            man += 0x100
        # overflow?
        if man >= self._den_upper:
            exp += 1
            man >>= 1
        # discard carry
        man ^= man & 0xff
        return exp, man, neg

    def _just_under(self):
        """Return the largest floating-point number less than the given value."""
        lexp, lman, lneg = self._denormalise()
        # decrease mantissa by one
        return self.new()._normalise(lexp, lman - 0x100, lneg)

    def _abs_gt_den(self, lden, rden):
        """Absolute value is greater than."""
        lexp, lman, _ = lden
        rexp, rman, _ = rden
        if lexp != rexp:
            return (lexp > rexp)
        return (lman > rman)

    def _div10_den(self, lden):
        """Divide by 10 in-place."""
        exp, man, neg = self._div_den(lden,
            self.new().from_bytes(self._ten)._denormalise())
        # perhaps this should be in _div_den
        while man < self._den_mask:
            exp -= 1
            man <<= 1
        return exp, man, neg

    def _mul10_den(self, den):
        """Multiply in-place by 10."""
        exp, man, neg = den
        # 10x == 2(x+4x)
        return self._add_den((exp+1, man, neg), (exp+3, man, neg))


    ##########################################################################
    # implementation: general

    _bias = None
    _shift = None
    _intformat = None
    _mask = None
    _posmask = None
    _signmask = None
    _den_mask = None
    _den_upper = None
    _carrymask = None

    def _denormalise(self):
        """Denormalise to shifted mantissa, exp, sign."""
        exp = bytearray(self._buffer)[-1]
        man = struct.unpack(
                self._intformat, b'\0' + bytearray(self._buffer)[:-1]
            )[0] | self._den_mask
        neg = self.is_negative()
        return exp, man, neg

    def _normalise(self, exp, man, neg):
        """Normalise from shifted mantissa, exp, sign."""
        # zero denormalised mantissa -> make zero
        if man == 0 or exp <= 0:
            self._buffer[:] = b'\0' * self.size
            return self
        # shift left if subnormal
        while man < (self._den_mask-1):
            exp -= 1
            man <<= 1
        # round to nearest; halves to even (Gaussian rounding)
        round_up = (man & 0xff > 0x80) or (man & 0xff == 0x80 and man & 0x100 == 0x100)
        man = (man & self._carrymask) + 0x100 * round_up
        if man >= self._den_upper:
            exp += 1
            man >>= 1
        # pack into byte representation
        struct.pack_into(
            self._intformat, self._buffer, 0, (man>>8) & (self._mask if neg else self._posmask)
        )
        if self._check_limits(exp, neg):
            self._buffer[-1:] = int2byte(exp)
        return self

    def _to_int_den(self):
        """Denormalised float to integer."""
        exp, man, neg = self._denormalise()
        exp -= self._bias
        if exp > 0:
            man <<= exp
        else:
            man >>= -exp
        return man, neg

    def _check_limits(self, exp, neg):
        """Overflow and underflow check
        returns:
            True if nonzero non-infinite
            False if underflow
        raises:
            OverflowError if overflow
        """
        if exp > 255:
            self.from_bytes(self.neg_max if neg else self.pos_max)
            raise OverflowError(self)
        elif exp <= 0:
            # set to zero, but leave mantissa as is
            self._buffer[-1:] = int2byte(0)
            return False
        return True

    def _bring_to_range(self, man, exp, lower, upper):
        """Bring mantissa to range (posmask, mask]."""
        while abs(man) <= lower:
            exp -= 1
            man <<= 1
        while abs(man) > upper:
            exp += 1
            man >>= 1
        return man, exp

    def _abs_gt(self, rhs):
        """Absolute values greater than."""
        # don't compare zeroes - failsafe, is not reached in code
        if self.is_zero(): # pragma: no cover
            return False
        rhscopy = bytearray(rhs._buffer)
        # so long as the sign is the same ...
        rhscopy[-2] &= (bytearray(self._buffer)[-2] | 0x7f)
        # ... we can compare floats as if they were ints
        for l, r in reversed(list(zip(bytearray(self._buffer), bytearray(rhscopy)))):
            if l > r:
                return True
            elif l < r:
                return False
        # equal
        return False

    def _add_den(self, lden, rden):
        """Denormalised add."""
        lexp, lman, lneg = lden
        rexp, rman, rneg = rden
        if rexp == 0:
            return lexp, lman, lneg
        if lexp == 0:
            return rexp, rman, rneg
        # ensure right is larger
        if lexp > rexp or (lexp == rexp and lman > rman):
            lexp, lman, lneg, rexp, rman, rneg = rexp, rman, rneg, lexp, lman, lneg
        # zero flag for quirky rounding
        # only set if all the bits we lose by matching exponents were zero
        zero_flag = lman & ((1<<(rexp-lexp))-1) == 0
        sub_flag = lneg != rneg
        # match exponents
        lman >>= (rexp - lexp)
        lexp = rexp
        # shortcut (this affects quirky rounding)
        if (lman < 0x80 or lman == 0x80 and zero_flag) and sub_flag:
            return rexp, rman, rneg
        # add mantissas, taking sign into account
        if not sub_flag:
            man, neg = lman + rman, lneg
            if man >= self._den_upper:
                lexp += 1
                man >>= 1
        else:
            man, neg = rman - lman, rneg
        # break tie for rounding if we're at exact half after dropping digits
        if not zero_flag and not sub_flag:
            man |= 0x1
        # attempt to match GW-BASIC subtraction rounding
        if sub_flag and (man & 0x1c0 == 0x80) and (man & 0x1df != 0x80):
            man &= (self._carrymask + 0x7f)
        return lexp, man, neg

    def _ipow_int(self, expt):
        """Raise to int power in-place."""
        # exponentiation by squares
        if expt < 0:
            self._ipow_int(-expt)
            self = self.new().from_bytes(self._one).idiv(self)
        elif expt > 1:
            if (expt % 2) == 0:
                self._ipow_int(expt // 2)
                self.imul(self)
            else:
                base = self.clone()
                self._ipow_int((expt-1) // 2)
                self.imul(self)
                self.imul(base)
        elif expt == 0:
            self = self.from_bytes(self._one)
        return self

    def _div_den(self, lden, rden):
        """Denormalised divide."""
        lexp, lman, lneg = lden
        rexp, rman, rneg = rden
        # signs
        lneg = (lneg != rneg)
        # subtract exponentials
        lexp -= rexp - self._bias - 8
        # long division of mantissas
        work_man = lman
        lman = 0
        lexp += 1
        while (rman > 0):
            lman <<= 1
            lexp -= 1
            if work_man > rman:
                work_man -= rman
                lman += 1
            rman >>= 1
        return lexp, lman, lneg


##############################################################################
# single-precision floating-point number

class Single(Float):
    """Single-precision MBF float."""

    sigil = b'!'
    size = 4

    exp_sign = b'E'
    digits = 7

    pos_max = b'\xff\xff\x7f\xff'
    neg_max = b'\xff\xff\xff\xff'

    _intformat = '<L'

    _bias = 128 + 24
    _shift = _bias - 129

    _den_mask = 0x80000000
    _den_upper = _den_mask * 2
    _carrymask = 0xffffff00

    _signmask = 0x800000
    _mask = 0xffffff
    _posmask = 0x7fffff

    _one = b'\x00\x00\x00\x81'
    _ten = b'\x00\x00\x20\x84'
    _lim_top = b'\x7f\x96\x18\x98' # 9999999, highest float less than 10e+7
    _lim_bot = b'\xff\x23\x74\x94' # 999999.9, highest float  less than 10e+6

    def to_token(self):
        """Return value as Single token."""
        return tk.T_SINGLE + self._buffer.tobytes()

    def from_token(self, token):
        """Set value to Single token."""
        if bytearray(token)[0] != ord(tk.T_SINGLE):
            raise ValueError('%s is not a Single token.' % repr(token))
        self._buffer[:] = token[-4:]
        return self

    def to_single(self):
        """Convert single to single (no-op)."""
        return self

    def to_double(self):
        """Convert single to double."""
        return Double(None, self._values).from_single(self)

    def to_float(self, allow_double=True):
        """Convert single to float."""
        return self


###############################################################################
# double-precision floating-point number

class Double(Float):
    """Double-precision MBF float."""

    sigil = b'#'
    size = 8

    exp_sign = b'D'
    digits = 16

    pos_max = b'\xff\xff\xff\xff\xff\xff\x7f\xff'
    neg_max = b'\xff\xff\xff\xff\xff\xff\xff\xff'

    _intformat = '<Q'

    _bias = 128 + 56
    _shift = _bias - 129

    _den_mask = 0x8000000000000000
    _den_upper = _den_mask * 2
    _carrymask = 0xffffffffffffff00

    _signmask = 0x80000000000000
    _mask = 0xffffffffffffff
    _posmask = 0x7fffffffffffff

    _one = b'\x00\x00\x00\x00\x00\x00\x00\x81'
    _ten = b'\x00\x00\x00\x00\x00\x00\x20\x84'
    _lim_top = b'\xff\xff\x03\xbf\xc9\x1b\x0e\xb6' # highest float less than 10e+16
    _lim_bot = b'\xff\xff\x9f\x31\xa9\x5f\x63\xb2' # highest float less than 10e+15

    def from_single(self, in_single):
        """Convert Single to Double in-place."""
        self._buffer[:4] = b'\0\0\0\0'
        self._buffer[4:] = in_single._buffer
        return self

    def to_token(self):
        """Return value as Single token."""
        return tk.T_DOUBLE + self._buffer.tobytes()

    def from_token(self, token):
        """Set value to Single token."""
        if bytearray(token)[0] != ord(tk.T_DOUBLE):
            raise ValueError('%s is not a Double token.' % repr(token))
        self._buffer[:] = token[-8:]
        return self

    def to_single(self):
        """Round double to single."""
        mybytes = self.to_bytes()
        single = Single(None, self._values).from_bytes(mybytes[4:])
        exp, man, neg = single._denormalise()
        # carry byte
        man += mybytes[3]
        return single._normalise(exp, man, neg)

    def to_double(self):
        """Convert double to double (no-op)."""
        return self

    def to_float(self, allow_double=True):
        """Convert double to float."""
        if allow_double:
            return self
        return self.to_single()


##############################################################################
# convert string representation to float

def str_to_decimal(s, allow_nonnum=True):
    """Return Float value for Python string."""
    found_sign, found_point, found_exp = False, False, False
    found_exp_sign, exp_neg, neg = False, False, False
    exp10, exponent, mantissa, digits, zeros = 0, 0, 0, 0, 0
    is_double, is_single = False, False
    for c in iterchar(s):
        # ignore whitespace throughout (x = 1   234  56  .5  means x=123456.5 in gw!)
        if c in BLANKS:
            continue
        if c in SEPARATORS:
            # ASCII separator chars invariably lead to zero result
            return False, 0, 0
        # determine sign
        if (not found_sign):
            found_sign = True
            # number has started; if no sign encountered here, sign must be pos.
            if c in b'+-':
                neg = (c == b'-')
                continue
        # parse numbers and decimal points, until 'E' or 'D' is found
        if (not found_exp):
            if b'0' <= c <= b'9':
                mantissa *= 10
                mantissa += ord(c) - ord(b'0')
                if found_point:
                    exp10 -= 1
                # keep track of precision digits
                if mantissa != 0:
                    digits += 1
                    if found_point and c == b'0':
                        zeros += 1
                    else:
                        zeros = 0
                continue
            elif c == b'.':
                found_point = True
                continue
            elif c.upper() in b'DE':
                found_exp = True
                is_double = (c.upper() == b'D')
                continue
            elif c == b'!':
                # makes it a single, even if more than eight digits specified
                is_single = True
                break
            elif c == b'#':
                is_double = True
                break
            else:
                if allow_nonnum:
                    break
                raise ValueError('Non-numerical character in string')
        # parse exponent
        elif (not found_exp_sign):
            # exponent has started; if no sign given, it must be pos.
            found_exp_sign = True
            if c in b'+-':
                exp_neg = (c == b'-')
                continue
        if (b'0' <= c <= b'9'):
            exponent *= 10
            exponent += ord(c) - ord(b'0')
            continue
        else:
            if allow_nonnum:
                break
            raise ValueError('Non-numerical character in string')
    if exp_neg:
        exp10 -= exponent
    else:
        exp10 += exponent
    # eight or more digits means double, unless single override
    if digits - zeros > 7 and not is_single:
        is_double = True
    return is_double, -mantissa if neg else mantissa, exp10

def _get_digits(mantissa, min_digits):
    """Get the digits for an int."""
    return (b'%d' % abs(mantissa)).rjust(min_digits, b'0')



"""
PC-BASIC - data

(c) 2021--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from ...compat import resources
import json

# copyright metadata
_METADATA = json.loads(resources.read_binary(__package__, 'meta.json'))
NAME, VERSION, AUTHOR, COPYRIGHT = (_METADATA[_key] for _key in (
    'name', 'version', 'author', 'copyright'
))

# release metadata, if available
try:
    _RELEASE_ID = json.loads(resources.read_binary(__package__, 'release.json'))
except EnvironmentError:
    _RELEASE_ID = {u'tag': u'', u'commit': u'unreleased', u'timestamp': u''}
TAG, TIMESTAMP, COMMIT = (_RELEASE_ID[_key] for _key in (u'tag', u'timestamp', u'commit'))
if COMMIT in TAG: # pragma: no cover
    LONG_VERSION = u'%s [%s %s]' % (VERSION, TAG, TIMESTAMP)
else: # pragma: no cover
    LONG_VERSION = u'%s [%s %s %s]' % (VERSION, TAG, COMMIT, TIMESTAMP)

# default font and codepage
DEFAULT_FONT = tuple(json.loads(resources.read_binary(__package__, 'font.json')))
DEFAULT_CODEPAGE = tuple(json.loads(resources.read_binary(__package__, 'codepage.json')))



"""
PC-BASIC - display.framebuffer
Emulated video memory

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import struct
import functools
import operator

from ...compat import xrange, int2byte, zip, iterbytes, PY2

from ..base import bytematrix


# video segment
MDA_SEGMENT = 0xb000
CGA_SEGMENT = 0xb800
EGA_SEGMENT = 0xa000


##############################################################################
# sprites & tiles

class PackedTileBuilder(object):
    """Packed-pixel (CGA) tiles."""

    def __init__(self, bits_per_pixel):
        """Initialise tile builder."""
        self._bitsperpixel = bits_per_pixel

    def __call__(self, pattern):
        """Build a flood-fill tile for CGA screens."""
        # in modes 1, (2), 3, 4, 5, 6 colours are encoded in consecutive bits
        # each byte represents one scan line
        return bytematrix.ByteMatrix.frompacked(
            pattern, height=len(pattern), items_per_byte=8//self._bitsperpixel
        )


class PlanedTileBuilder(object):
    """Interlaced-plane (EGA) tiles."""

    def __init__(self, number_planes):
        """Initialise sprite builder."""
        # number of colour planes
        self._number_planes = number_planes

    def __call__(self, pattern):
        """Build a flood-fill tile."""
        # append nulls until we can cleanly partition into planes
        extra_chars = len(pattern) % self._number_planes
        if extra_chars:
            pattern.extend(bytearray(self._number_planes - extra_chars))
        # unpack bytes into pattern
        allplanes = bytematrix.ByteMatrix.frompacked(
            pattern, height=len(pattern), items_per_byte=8
        )
        planes = (
            allplanes[_plane::self._number_planes, :] << _plane
            for _plane in range(self._number_planes)
        )
        tile = functools.reduce(operator.__ior__, planes)
        return tile


class PackedSpriteBuilder(object):
    """Packed-pixel (CGA) sprite builder."""

    # no width quirks
    width_factor = 1

    def __init__(self, bits_per_pixel):
        self._bitsperpixel = bits_per_pixel

    def pack(self, sprite):
        """Pack the sprite into bytearray."""
        # sprite size record
        size_record = struct.pack('<HH', sprite.width * self._bitsperpixel, sprite.height)
        # interval_to_bytes
        packed = sprite.packed(items_per_byte=8 // self._bitsperpixel)
        return size_record + packed

    def unpack(self, array):
        """Unpack bytearray into sprite."""
        row_bits, height = struct.unpack('<HH', array[0:4])
        width = row_bits // self._bitsperpixel
        row_bytes = (width * self._bitsperpixel + 7) // 8
        byte_size = row_bytes * height
        # bytes_to_interval
        packed = array[4:4+byte_size]
        # ensure iterations over memoryview yield int, not bytes, in Python 2
        # frompacked can't take interators, would need a width argument
        #packed = iterbytes(packed)
        if PY2: # pragma: no cover
            packed = bytearray(packed)
        sprite = bytematrix.ByteMatrix.frompacked(
            packed, height, items_per_byte=8 // self._bitsperpixel
        )
        # clip to requested width
        sprite = sprite[:, :width]
        return sprite


class PlanedSpriteBuilder(object):
    """Sprite builder with interlaced colour planes (EGA sprites)."""

    # no width quirks
    width_factor = 1

    # ** byte mapping for sprites in EGA modes
    # sprites have 8 pixels per byte
    # with colour planes in consecutive rows
    # each new row is aligned on a new byte

    def __init__(self, number_planes):
        """Initialise sprite builder."""
        # number of colour planes
        self._number_planes = number_planes

    def pack(self, sprite):
        """Pack the sprite into bytearray."""
        # extract colour planes
        # note that to get the plane this should be bit-masked - (s >> _p) & 1
        # but bytematrix.packbytes will do this for us
        sprite_planes = (
            (sprite >> _plane)  # & 1
            for _plane in range(self._number_planes)
        )
        # pack the bits into bytes
        #interval_to_bytes
        packed_planes = list(
            _sprite.packed(items_per_byte=8)
            for _sprite in sprite_planes
        )
        # interlace row-by-row
        row_bytes = (sprite.width + 7) // 8
        length = sprite.height * self._number_planes * row_bytes
        interlaced = bytearray().join(
            _packed[_row_offs : _row_offs+row_bytes]
            for _row_offs in range(0, length, row_bytes)
            for _packed in packed_planes
        )
        size_record = struct.pack('<HH', sprite.width, sprite.height)
        return size_record + interlaced

    def unpack(self, array):
        """Build sprite from bytearray in EGA modes."""
        width, height = struct.unpack('<HH', array[:4])
        row_bytes = (width + 7) // 8
        length = height * self._number_planes * row_bytes
        packed = array[4:4+length]
        # ensure iterations over memoryview yield int, not bytes, in Python 2
        packed = iterbytes(packed)
        # unpack all planes
        #bytes_to_interval
        allplanes = bytematrix.ByteMatrix.frompacked(
            packed, height=height*self._number_planes, items_per_byte=8
        )
        # clip to requested width
        allplanes = allplanes[:, :width]
        # de-interlace planes
        sprite_planes = (
            allplanes[_plane::self._number_planes, :] << _plane
            for _plane in range(self._number_planes)
        )
        # combine planes
        sprite = functools.reduce(operator.__ior__, sprite_planes)
        return sprite


class Tandy6SpriteBuilder(PlanedSpriteBuilder):
    """Wrapper for Tandy mode 6's sprite-size record quirks."""

    # width is twice what's recorded
    width_factor = 2

    def pack(self, sprite):
        """Pack sprite, report only half the width packed."""
        record = PlanedSpriteBuilder.pack(self, sprite)
        width = sprite.width // self.width_factor
        width_record = struct.pack('<H', width)
        return width_record + record[2:]

    def unpack(self, array):
        """Unpack sprite, twice the width reported."""
        width, = struct.unpack('<H', array[:2])
        width *= self.width_factor
        size_record = struct.pack('<H', width)
        # explicit conversion to bytes only needed for Python 2, in Python 3 concatenation is ok
        # via bytearray as otherwide it gives str representation in Python 2
        return PlanedSpriteBuilder.unpack(self, size_record + bytes(bytearray(array[2:])))


##############################################################################
# framebuffer memory map

class _MemoryMapper(object):
    """Map between coordinates and locations in the framebuffer."""

    def __init__(
            self, video_mem_size, max_pages, page_size
        ):
        """Initialise video mode settings."""
        # override this
        self._video_segment = None
        self._page_size = page_size
        self._video_mem_size = video_mem_size
        self._max_pages = max_pages

    def set_video_mem_size(self, new_size):
        """Set available video memory."""
        self._video_mem_size = new_size

    @property
    def num_pages(self):
        """Number of available pages."""
        num_pages = self._video_mem_size // self._page_size
        if self._max_pages:
            return min(self._max_pages, num_pages)
        return num_pages

    @property
    def page_size(self):
        """Size in bytes of video page."""
        return self._page_size

    def get_all_memory(self, display):
        """Obtain a copy of all video memory."""
        addr = self._video_segment * 0x10
        buffer = self.get_memory(display, addr, self._video_mem_size)
        return addr, buffer

    def get_memory(self, display, addr, num_bytes):
        """Retrieve bytes from video memory, stub."""

    def set_memory(self, display, addr, bytes):
        """Set bytes in video memory, stub."""


class TextMemoryMapper(_MemoryMapper):
    """Map between coordinates and locations in the textmode framebuffer."""

    def __init__(self, text_height, text_width, video_mem_size, max_pages, is_mono):
        """Initialise video mode settings."""
        page_size = 0x1000 if text_width == 80 else 0x800
        _MemoryMapper.__init__(self, video_mem_size, max_pages, page_size)
        self._video_segment = MDA_SEGMENT if is_mono else CGA_SEGMENT
        self._text_height = text_height
        self._text_width = text_width

    def get_memory(self, display, addr, num_bytes):
        """Retrieve bytes from textmode video memory."""
        addr -= self._video_segment * 0x10
        mem_bytes = bytearray(num_bytes)
        for i in xrange(num_bytes):
            page, offset = divmod(addr+i, self._page_size)
            row, row_offset = divmod(offset, self._text_width*2)
            col = row_offset // 2
            try:
                if (addr+i) % 2:
                    mem_bytes[i] = display.pages[page].get_attr(1 + row, 1 + col)
                else:
                    mem_bytes[i] = display.pages[page].get_byte(1 + row, 1 + col)
            except IndexError:
                pass
        return mem_bytes

    def set_memory(self, display, addr, mem_bytes):
        """Set bytes in textmode video memory."""
        addr -= self._video_segment*0x10
        last_row = 0
        for i in xrange(len(mem_bytes)):
            page, offset = divmod(addr+i, self._page_size)
            row, row_offset = divmod(offset, self._text_width*2)
            col = row_offset // 2
            try:
                if (addr+i) % 2:
                    c = display.pages[page].get_byte(1+row, 1+col)
                    a = mem_bytes[i]
                else:
                    c = mem_bytes[i]
                    a = display.pages[page].get_attr(1+row, 1+col)
                display.pages[page].put_char_attr(1+row, 1+col, int2byte(c), a)
            except IndexError:
                pass
            last_row = 1 + row


class GraphicsMemoryMapper(_MemoryMapper):
    """Map between coordinates and locations in the graphical framebuffer."""

    def __init__(
            self, pixel_height, pixel_width,
            video_mem_size, max_pages, interleave_times, bank_size, bitsperpixel
        ):
        """Initialise video mode settings."""
        page_size = interleave_times * bank_size
        _MemoryMapper.__init__(self, video_mem_size, max_pages, page_size)
        self._pixel_height = pixel_height
        self._pixel_width = pixel_width
        # cga bank_size = 0x2000 interleave_times=2
        self._interleave_times = interleave_times
        self._bank_size = bank_size
        self._bitsperpixel = bitsperpixel
        # number of pixels referenced in each byte of a plane
        self._ppb = 8 // bitsperpixel
        # strides
        self._bytes_per_row = self._pixel_width * bitsperpixel // 8

    def _get_coords(self, addr):
        """Get video page and coordinates for address."""
        # override
        return 0, 0, 0

    def _coord_ok(self, page, x, y):
        """Check if a page and coordinates are within limits."""
        return (
            page >= 0 and page < self.num_pages and
            x >= 0 and x < self._pixel_width and
            y >= 0 and y < self._pixel_height
        )

    def set_plane(self, plane):
        """Set the current colour plane (EGA only)."""

    def set_plane_mask(self, mask):
        """Set the current colour plane mask (EGA only)."""

    def _walk_memory(self, addr, num_bytes, factor=1):
        """Iterate over graphical memory (pixel-by-pixel, contiguous rows)."""
        # factor supports tandy-6 mode, which has 8 pixels per 2 bytes
        # with alternating planes in even and odd bytes (i.e. ppb==8)
        ppb = factor * self._ppb
        page_size = self._page_size // factor
        bank_size = self._bank_size // factor
        row_size = self._bytes_per_row // factor
        # first row
        page, x, y = self._get_coords(addr)
        offset = min(row_size - x//ppb, num_bytes)
        if self._coord_ok(page, x, y):
            yield page, x, y, 0, offset
        # full rows
        bank_offset, page_offset, start_y = 0, 0, y
        while page_offset + bank_offset + offset < num_bytes:
            y += self._interleave_times
            # not an integer number of rows in a bank
            if offset >= bank_size:
                bank_offset += bank_size
                start_y += 1
                offset, y = 0, start_y
                if bank_offset >= page_size:
                    page_offset += page_size
                    page += 1
                    bank_offset, offset = 0, 0
                    y, start_y = 0, 0
            if self._coord_ok(page, 0, y):
                ofs = page_offset + bank_offset + offset
                if ofs + row_size > num_bytes:
                    yield page, 0, y, ofs, num_bytes - ofs
                else:
                    yield page, 0, y, ofs, row_size
            offset += row_size


class CGAMemoryMapper(GraphicsMemoryMapper):
    """Map between coordinates and locations in the CGA framebuffer."""

    def __init__(
            self, pixel_height, pixel_width,
            video_mem_size, max_pages, interleave_times, bank_size, bitsperpixel
        ):
        """Initialise video mode settings."""
        GraphicsMemoryMapper.__init__(
            self, pixel_height, pixel_width,
            video_mem_size, max_pages, interleave_times, bank_size, bitsperpixel
        )
        self._video_segment = CGA_SEGMENT

    def _get_coords(self, addr):
        """Get video page and coordinates for address."""
        addr = int(addr) - self._video_segment * 0x10
        # modes 1-5: interleaved scan lines, pixels sequentially packed into bytes
        page, addr = divmod(addr, self._page_size)
        # 2 x interleaved scan lines of 80bytes
        bank, offset = divmod(addr, self._bank_size)
        row, col = divmod(offset, self._bytes_per_row)
        x = col * 8 // self._bitsperpixel
        y = bank + self._interleave_times * row
        return page, x, y

    def set_memory(self, display, addr, byte_array):
        """Set bytes in CGA memory."""
        for page, x, y, ofs, length in self._walk_memory(addr, len(byte_array)):
            #bytes_to_interval
            pixarray = bytematrix.ByteMatrix.frompacked(
                byte_array[ofs:ofs+length], height=1, items_per_byte=self._ppb
            )
            display.pages[page].pixels[y, x:x+pixarray.width] = pixarray

    def get_memory(self, display, addr, num_bytes):
        """Retrieve bytes from CGA memory."""
        byte_array = bytearray(num_bytes)
        for page, x, y, ofs, length in self._walk_memory(addr, num_bytes):
            #interval_to_bytes
            pixarray = display.pages[page].pixels[y, x:x+length*self._ppb]
            byte_array[ofs:ofs+length] = pixarray.packed(self._ppb)
        return byte_array


class EGAMemoryMapper(GraphicsMemoryMapper):
    """Map between coordinates and locations in the EGA framebuffer."""

    def __init__(
            self, pixel_height, pixel_width,
            video_mem_size, max_pages, interleave_times, bank_size, bitsperpixel
        ):
        """Initialise video mode settings."""
        GraphicsMemoryMapper.__init__(
            self, pixel_height, pixel_width,
            video_mem_size, max_pages, interleave_times, bank_size, bitsperpixel
        )
        self._video_segment = EGA_SEGMENT
        # EGA uses colour planes, 1 bpp for each plane
        # this is used by walk_memory to determine strides
        self._ppb = 8
        self._bytes_per_row = self._pixel_width // 8
        self._planes_used = range(4)
        # additional colour plane mask
        self._master_plane_mask = 0x07
        # current ega memory colour plane to read
        self._plane = 0
        # current ega memory colour planes to write to
        self._plane_mask = 0xff

    def set_planes_used(self, planes_used):
        """EGA specific settings."""
        self._planes_used = planes_used
        # additional colour plane mask
        self._master_plane_mask = sum(2**_plane for _plane in planes_used)

    def set_plane(self, plane):
        """Set the current colour plane."""
        self._plane = plane

    def set_plane_mask(self, mask):
        """Set the current colour plane mask."""
        self._plane_mask = mask

    def _get_coords(self, addr):
        """Get video page and coordinates for address."""
        addr = int(addr) - self._video_segment * 0x10
        # modes 7-9: 1 bit per pixel per colour plane
        page, addr = divmod(addr, self._page_size)
        row, row_offs = divmod(addr, self._bytes_per_row)
        x, y = row_offs*8, row
        return page, x, y

    def get_memory(self, display, addr, num_bytes):
        """Retrieve bytes from EGA memory."""
        plane = self._plane % (max(self._planes_used) + 1)
        byte_array = bytearray(num_bytes)
        if plane not in self._planes_used:
            return byte_array
        for page, x, y, ofs, length in self._walk_memory(addr, num_bytes):
            pixarray = display.pages[page].pixels[y, x:x+length*8]
            byte_array[ofs:ofs+length] = (pixarray >> plane).packed(8)
        return byte_array

    def set_memory(self, display, addr, byte_array):
        """Set bytes in EGA video memory."""
        # EGA memory is planar with memory-mapped colour planes.
        # Within a plane, 8 pixels are encoded into each byte.
        # The colour plane is set through a port OUT and
        # determines which bit of each pixel's attribute is affected.
        mask = self._plane_mask & self._master_plane_mask
        # return immediately for unused colour planes
        if mask == 0:
            return
        for page, x, y, ofs, length in self._walk_memory(addr, len(byte_array)):
            pixarray = (
                bytematrix.ByteMatrix.frompacked(
                    byte_array[ofs:ofs+length], height=1, items_per_byte=8
                ).render(0, mask)
            )
            width = pixarray.width
            substrate = display.pages[page].pixels[y, x:x+width] & ~mask
            display.pages[page].pixels[y, x:x+width] = (pixarray & mask) | substrate


class Tandy6MemoryMapper(GraphicsMemoryMapper):
    """Map between coordinates and locations in the Tandy SCREEN 6 framebuffer."""

    def __init__(
            self, pixel_height, pixel_width,
            video_mem_size, max_pages, interleave_times, bank_size, bitsperpixel
        ):
        """Initialise video mode settings."""
        GraphicsMemoryMapper.__init__(
            self, pixel_height, pixel_width,
            video_mem_size, max_pages, interleave_times, bank_size, bitsperpixel
        )
        self._video_segment = CGA_SEGMENT
        # mode 6: 4x interleaved scan lines, 8 pixels per two bytes,
        # low attribute bits stored in even bytes, high bits in odd bytes.
        self._bytes_per_row = self._pixel_width * 2 // 8

    def _get_coords(self, addr):
        """Get video page and coordinates for address."""
        addr =  int(addr) - self._video_segment * 0x10
        page, addr = divmod(addr, self._page_size)
        # 4 x interleaved scan lines of 160bytes
        bank, offset = divmod(addr, self._bank_size)
        row, col = divmod(offset, self._bytes_per_row)
        x = (col // 2) * 8
        y = bank + 4 * row
        return page, x, y

    def get_memory(self, display, addr, num_bytes):
        """Retrieve bytes from Tandy 640x200x4 """
        # 8 pixels per 2 bytes
        # low attribute bits stored in even bytes, high bits in odd bytes.
        half_len = (num_bytes+1) // 2
        hbytes = bytearray(half_len), bytearray(half_len)
        for parity, byte_array in enumerate(hbytes):
            plane = parity ^ (addr % 2)
            for page, x, y, ofs, length in self._walk_memory(addr, num_bytes, 2):
                pixarray = display.pages[page].pixels[y, x : x + length*self._ppb*2]
                #hbytes[parity][ofs:ofs+length] = interval_to_bytes(pixarray, self._ppb*2, plane)
                byte_array[ofs:ofs+length] = (pixarray >> plane).packed(self._ppb * 2)
        # resulting array may be too long by one byte, so cut to size
        return [_item for _pair in zip(*hbytes) for _item in _pair] [:num_bytes]

    def set_memory(self, display, addr, byte_array):
        """Set bytes in Tandy 640x200x4 memory."""
        hbytes = byte_array[0::2], byte_array[1::2]
        # Tandy-6 encodes 8 pixels per byte, alternating colour planes.
        # I.e. even addresses are 'colour plane 0', odd ones are 'plane 1'
        for parity, half in enumerate(hbytes):
            plane = parity ^ (addr % 2)
            mask = 2 ** plane
            for page, x, y, ofs, length in self._walk_memory(addr, len(byte_array), 2):
                #pixarray = bytes_to_interval(hbytes[parity][ofs:ofs+length], 2*self._ppb, mask)
                pixarray = (
                    bytematrix.ByteMatrix.frompacked(
                        # what's the deal with the empty bytearrays here in some of the tests?
                        half[ofs:ofs+length], height=1, items_per_byte=2*self._ppb
                    ) << plane
                )
                width = pixarray.width
                substrate = display.pages[page].pixels[y, x:x+width] & ~mask
                display.pages[page].pixels[y, x:x+width] = (pixarray & mask) | substrate



"""
PC-BASIC - compat.posix_console
POSIX console support with ANSI escape sequences

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

# pylint: disable=no-name-in-module

import os
import io
import sys
import tty
import time
import termios
import select
import fcntl
import array
import struct
import atexit
import locale
import logging
from collections import deque
from contextlib import contextmanager
try:
    import curses
except ImportError:
    curses = None

from .base import MACOS, PY2, HOME_DIR
from .streams import StdIOBase


# ANSI escape codes
# for reference, see:
# http://en.wikipedia.org/wiki/ANSI_escape_code
# http://misc.flogisoft.com/bash/tip_colors_and_formatting
# http://www.termsys.demon.co.uk/vtansi.htm
# http://ascii-table.com/ansi-escape-sequences-vt-100.php
# https://invisible-island.net/xterm/ctlseqs/ctlseqs.html


# ANSI escape sequences
# - these are supported by xterm, gnome-terminal
# - konsole and Terminal.app ignore the palette sequences
# - konsole (pre 18.08) breaks on the cursor shape sequence
# - Terminal.app ignores cursor shape but does not break
# - we use the delete/insert lines sequences rather than scroll as they are better supported
# unfortunately terminfo is spotty on cursor shape and palette functionality,
# plus most consoles claim to be xterm anyway

if os.getenv('TERM', default='').startswith('linux'):
    # linux framebuffer console
    ANSI_OVERRIDES = dict(
        # 1 invisible 2 line 3 third 4 half block 5 two thirds 6 full block
        # https://linuxgazette.net/137/anonymous.html
        _cursor_block = b'\x1b[?4c',
        _cursor_line = b'\x1b[?2c',
        _reset_cursor = b'\x1b[?0c',
    )
else:
    # xterm and family
    ANSI_OVERRIDES = dict(
        # 1 blinking block 2 block 3 blinking line 4 line
        _cursor_block = b'\x1b[1 q', # Ss 1 ?
        _cursor_line = b'\x1b[3 q', # Ss 3 ?
        # reset colour and shape
        _reset_cursor = b'\x1b]112\a\x1b[1 q',
        # follow the format of initc
        # Cs ?
        _cursor_color = b'\x1b]12;#%p1%{255}%*%{1000}%/%2.2X%p2%{255}%*%{1000}%/%2.2X%p3%{255}%*%{1000}%/%2.2X\a',
        # window properties
        _resize = b'\x1b[8;%p1%d;%p2%d;t', ## ?
        # status line (caption)
        tsl = b'\x1b]2;',
        fsl = b'\a',
    )


# input key codes
###################################################################################################

# terminfo has only a few keycodes
# fortunately we can just include all codes for all systems
# as there are no conflicting definitions

# xterm keys that support modifiers
_MOD_PATTERNS = {
    u'\x1b[11%s~': 'F1',
    u'\x1b[12%s~': 'F2',
    u'\x1b[13%s~': 'F3',
    u'\x1b[14%s~': 'F4',
    u'\x1b[15%s~': 'F5',
    u'\x1b[17%s~': 'F6',
    u'\x1b[18%s~': 'F7',
    u'\x1b[19%s~': 'F8',
    u'\x1b[20%s~': 'F9',
    u'\x1b[21%s~': 'F10',
    u'\x1b[23%s~': 'F11',
    u'\x1b[24%s~': 'F12',
    u'\x1b[1%sF': 'END',
    u'\x1b[1%sH': 'HOME',
    u'\x1b[1%sA': 'UP',
    u'\x1b[1%sB': 'DOWN',
    u'\x1b[1%sC': 'RIGHT',
    u'\x1b[1%sD': 'LEFT',
    u'\x1b[2%s~': 'INSERT',
    u'\x1b[3%s~': 'DELETE',
    u'\x1b[5%s~': 'PAGEUP',
    u'\x1b[6%s~': 'PAGEDOWN',
}

# xterm modifier codes
_MOD_CODES = {
    u'': set(),
    u';2': {'SHIFT'},
    u';3': {'ALT'},
    u';4': {'SHIFT', 'ALT'},
    u';5': {'CTRL'},
    u';6': {'SHIFT', 'CTRL'},
    u':7': {'CTRL', 'ALT'},
    u';8': {'SHIFT', 'CTRL', 'ALT'},
}

# construct ansi to output mapping for xterm codes
ANSI_TO_KEYMOD = {
    pattern % (modcode,): (key, mods)
    for modcode, mods in _MOD_CODES.items()
    for pattern, key in _MOD_PATTERNS.items()
}

# unmodified keys
_UNMOD_KEYS = {
    # used by the linux framebuffer console
    # also, \e[25~ is shift+F1, etc
    u'\x1b[[A': 'F1',
    u'\x1b[[B': 'F2',
    u'\x1b[[C': 'F3',
    u'\x1b[[D': 'F4',
    u'\x1b[[E': 'F5',
    u'\x1b[4~': 'END',
    u'\x1b[1~': 'HOME',
    # CSI-based key codes (without the number 1)
    u'\x1b[F': 'END',
    u'\x1b[H': 'HOME',
    u'\x1b[A': 'UP',
    u'\x1b[B': 'DOWN',
    u'\x1b[C': 'RIGHT',
    u'\x1b[D': 'LEFT',
    # SS3-based key codes (used by xterm in smkx mode)
    u'\x1bOP': 'F1',
    u'\x1bOQ': 'F2',
    u'\x1bOR': 'F3',
    u'\x1bOS': 'F4',
    u'\x1bOF': 'END',
    u'\x1bOH': 'HOME',
    u'\x1bOA': 'UP',
    u'\x1bOB': 'DOWN',
    u'\x1bOC': 'RIGHT',
    u'\x1bOD': 'LEFT',
}
ANSI_TO_KEYMOD.update({sequence: (key, set()) for sequence, key in _UNMOD_KEYS.items()})

# shifted keys
_SHIFT_KEYS = {
    # shifted F-keys used by the linux framebuffer console
    u'\x1b[25~': 'F1',
    u'\x1b[26~': 'F2',
    u'\x1b[28~': 'F3',
    u'\x1b[29~': 'F4',
    u'\x1b[31~': 'F5',
    u'\x1b[32~': 'F6',
    u'\x1b[33~': 'F7',
    u'\x1b[34~': 'F8',
    # xterm shift+TAB
    u'\x1b[[Z': 'TAB',
}
ANSI_TO_KEYMOD.update({sequence: (key, {'SHIFT'}) for sequence, key in _SHIFT_KEYS.items()})

# keypad codes with numlock off
# arrow keys, ins, del etc already included
# u'\x1bOE': keypad 5
# u'\x1bOM': keypad Enter
# u'\x1bOk': keypad +
# u'\x1bOm': keypad -
# u'\x1bOj': keypad *
# u'\x1bOo': keypad /

# esc + char means alt+key; lowercase
ANSI_TO_KEYMOD.update({u'\x1b%c' % (c + 32,): (chr(c + 32), {'ALT'}) for c in range(65, 91)})
# uppercase
ANSI_TO_KEYMOD.update({u'\x1b%c' % (c,): (chr(c + 32), {'ALT', 'SHIFT'}) for c in range(65, 91)})
# digits, controls & everything else
ANSI_TO_KEYMOD.update({u'\x1b%c' % (c,): (chr(c), {'ALT'}) for c in range(0, 65)})
ANSI_TO_KEYMOD.update({u'\x1b%c' % (c,): (chr(c), {'ALT'}) for c in range(91, 128)})


# colour palettes
###################################################################################################

# mapping of the first 8 attributes of the default CGA palette
# so that non-RGB terminals use sensible colours
# black, blue, green, cyan, red, magenta, yellow, white
EGA_TO_ANSI = (0, 4, 2, 6, 1, 5, 3, 7)

# default palette - these are in fact the 16 CGA colours
# this gets overwritten anyway
DEFAULT_PALETTE = (
    (0x00, 0x00, 0x00), (0x00, 0x00, 0xaa), (0x00, 0xaa, 0x00), (0x00, 0xaa, 0xaa),
    (0xaa, 0x00, 0x00), (0xaa, 0x00, 0xaa), (0xaa, 0x55, 0x00), (0xaa, 0xaa, 0xaa),
    (0x55, 0x55, 0x55), (0x55, 0x55, 0xff), (0x55, 0xff, 0x55), (0x55, 0xff, 0xff),
    (0xff, 0x55, 0x55), (0xff, 0x55, 0xff), (0xff, 0xff, 0x55), (0xff, 0xff, 0xff)
)


# implementation
###################################################################################################


class StdIO(StdIOBase):
    """Holds standard unicode streams."""

    if PY2: # pragma: no cover
        def _attach_stdin(self):
            self.stdin = self._wrap_input_stream(sys.stdin)

        def _attach_output_stream(self, stream_name, redirected=False):
            stream = getattr(sys, '__%s__' % (stream_name,))
            new_stream = self._wrap_output_stream(stream)
            setattr(self, stream_name, new_stream)


def init_stdio():
    """Platform-specific initialisation."""
    # set locale - this is necessary for curses and *maybe* for clipboard handling
    # there's only one locale setting so best to do it all upfront here
    # NOTE that this affects str.upper() etc.
    # don't do this on Windows as it makes the console codepage different from the stdout encoding ?
    try:
        locale.setlocale(locale.LC_ALL, '')
    except locale.Error as e:
        # mis-configured locale can throw an error here, no need to crash
        logging.error(e)


# output buffer for ioctl call
_sock_size = array.array('i', [0])


class PosixConsole(object):
    """POSIX-based console implementation."""

    def __init__(self, stdio):
        """Set up the console."""
        # buffer to save termios state
        if not sys.stdin.isatty() or not sys.stdout.isatty():
            raise EnvironmentError('Not a terminal')
        self._term_attr = termios.tcgetattr(sys.stdin.fileno())
        self._stdio = stdio
        # preserve original terminal size
        self._orig_size = self.get_size()
        self._height, _ = self._orig_size
        # input buffer
        self._read_buffer = deque()
        # palette
        self._palette = list(DEFAULT_PALETTE)
        # needed to access curses.tiget* functions
        if curses:
            curses.setupterm()
        self._muffle = None

    ##########################################################################
    # terminal modes

    def get_size(self):
        """Get terminal size."""
        try:
            return struct.unpack(
                'HHHH', fcntl.ioctl(sys.stdout.fileno(), termios.TIOCGWINSZ, b'\0'*8)
            )[:2]
        except Exception:
            return 25, 80

    def set_raw(self):
        """Enter raw terminal mode (no echo, don't exit on ctrl-C, ...)."""
        tty.setraw(sys.stdin.fileno())

    def unset_raw(self):
        """Leave raw terminal mode."""
        termios.tcsetattr(sys.stdin.fileno(), termios.TCSADRAIN, self._term_attr)


    ##########################################################################
    # ansi output

    def start_screen(self):
        """Enter full-screen/application mode."""
        # suppress stderr to avoid log messages defacing the application screen
        self._muffle = self._stdio.pause('stderr')
        self._muffle.__enter__()  # pylint: disable=no-member
        self.set_raw()
        # switch to alternate buffer
        self._emit_ti('smcup')
        # set application keypad / keypad transmit mode
        self._emit_ti('smkx')

    def close_screen(self):
        """Leave full-screen/application mode."""
        self.reset()
        self._emit_ti('rmkx')
        if not self._emit_ti('rmcup'):
            self._emit_ti('clear')
        self.unset_raw()
        if self._muffle is not None:
            self._muffle.__exit__(None, None, None)  # pylint: disable=no-member

    def reset(self):
        """Reset to defaults."""
        self._emit_ti('oc')
        self._emit_ti('op')
        self._emit_ti('sgr0')
        self._emit_ti('cnorm')
        self._emit_ti('_reset_cursor')
        self._emit_ti('_resize', *self._orig_size)

    def write(self, unicode_str):
        """Write (unicode) text to console."""
        self._stdio.stdout.write(unicode_str)
        self._stdio.stdout.flush()

    def _emit_ti(self, capability, *args):
        """Emit escape code."""
        if not curses:
            return False
        try:
            pattern = ANSI_OVERRIDES[capability]
        except KeyError:
            pattern = curses.tigetstr(capability)
        if pattern:
            ansistr = curses.tparm(pattern, *args).decode('ascii')
            self._stdio.stdout.write(ansistr)
            self._stdio.stdout.flush()
            return True
        return False

    def set_caption(self, caption):
        """Set terminal caption."""
        if self._emit_ti('tsl'):
            self._stdio.stdout.write(caption)
            self._emit_ti('fsl')

    def resize(self, height, width):
        """Resize terminal."""
        self._emit_ti('_resize', height, width)
        self._height = height
        # start below the current output
        self.clear()

    def clear(self):
        """Clear the screen and home the cursor."""
        self._emit_ti('clear')

    def clear_row(self, width=None):
        """Clear the current row."""
        if width is None:
            self._emit_ti('cr')
            self._emit_ti('el')
        else:
            self._emit_ti('sc')
            self._emit_ti('hpa', width-1)
            self._emit_ti('el1')
            self._emit_ti('rc')

    def show_cursor(self, block=False):
        """Show the cursor."""
        self._emit_ti('cnorm')
        if block:
            self._emit_ti('_cursor_block')
        else:
            self._emit_ti('_cursor_line')

    def hide_cursor(self):
        """Hide the cursor."""
        self._emit_ti('civis')

    def move_cursor_to(self, row, col):
        """Move cursor to a new position."""
        self._emit_ti('hpa', col-1)
        self._emit_ti('vpa', row-1)

    def scroll(self, top, bottom, rows):
        """Scroll the region between top and bottom one row up (-) or down (+)."""
        if bottom > top:
            self._emit_ti('csr', top-1, bottom-1)
            self._emit_ti('hpa', 0)
            self._emit_ti('vpa', top-1)
            if rows < 0:
                self._emit_ti('dl', -rows)
            elif rows > 0:
                self._emit_ti('il', rows)
            self._emit_ti('csr', 0, self._height-1)

    def set_cursor_colour(self, colour):
        """Set the current cursor colour attribute."""
        try:
            red, green, blue = self._palette[colour]
            self._emit_ti('_cursor_color', (red*1000)//255, (green*1000)//255, (blue*1000)//255)
        except KeyError:
            pass

    def set_attributes(self, fore, back, blink, underline):
        """Set current attributes."""
        # use "bold" ANSI colours for the upper 8 EGA attributes
        self._emit_ti('sgr0')
        self._emit_ti('setaf', 8 * (fore // 8) + EGA_TO_ANSI[fore % 8])
        self._emit_ti('setab', EGA_TO_ANSI[back])
        if blink:
            self._emit_ti('blink')
        if underline:
            self._emit_ti('smul')

    def set_palette_entry(self, attr, red, green, blue):
        """Set palette entry for attribute (0--16)."""
        # keep a record, mainly for cursor colours
        self._palette[attr] = red, green, blue
        # set the ANSI palette
        ansi_attr = 8*(attr//8) + EGA_TO_ANSI[attr%8]
        self._emit_ti('initc', ansi_attr, (red*1000)//255, (green*1000)//255, (blue*1000)//255)

    ##########################################################################
    # input

    def key_pressed(self):
        """Return whether a character is ready to be read from the keyboard."""
        return select.select([sys.stdin], [], [], 0)[0] != []

    def read_key(self):
        """
        Read keypress from console. Non-blocking.
        Returns tuple (unicode, keycode, set of mods)
        """
        sequence = read_all_available(self._stdio.stdin)
        if sequence is None:
            # stream closed, send ctrl-d
            return u'\x04', 'd', {'CTRL'}
        elif not sequence:
            return u'', None, set()
        # ansi sequences start with ESC (\x1b), but let ESC by itself through
        if len(sequence) > 1 and sequence[0] == u'\x1b':
            # drop unrecognised sequences
            key, mod = ANSI_TO_KEYMOD.get(sequence, (u'', ()))
            return u'', key, mod
        else:
            return sequence, None, set()


def _is_console_app():
    """To see if we are a console app, check if we can treat stdin like a tty, file or socket."""
    if not sys.stdin.isatty():
        try:
            fcntl.ioctl(sys.stdin, termios.FIONREAD, _sock_size)
        except EnvironmentError:
            # maybe /dev/null, but not a real file or console
            return False
    return True


IS_CONSOLE_APP = _is_console_app()




def read_all_available(stream):
    """Read all available bytes or unicode from a stream; nonblocking; None if closed."""
    try:
        # select: check if buffer has characters/lines to read
        # NOTE - select call works for files & sockets & character devices on unix; sockets only on Windows
        has_bytes_available = select.select([stream], [], [], 0)[0]
    except io.UnsupportedOperation:
        # select needs an actual file or socket that has a fileno, BytesIO, StringIO etc not supported
        return stream.read() or None
    try:
        # select gives bytes counts for unicode streams which is pretty useless
        # so if provided with a unicode stream, take the buffer and decode back to unicode ourselves
        encoding = stream.encoding
        stream = stream.buffer
    except AttributeError as exception:
        encoding = None
        # we might still have a unicode stream it just doesn't have the attributes
        # raise an error here, otherwise we might try to read too many chars and block later
        # we need to check type as b'' == u'' in python2
        if type(stream.read(0)) == type(u''):
            raise TypeError(
                "Can't perform non-blocking read from this text stream: %s"
                % (exception,)
            )
    if not has_bytes_available:
        # nothing currently available to read.
        # return an empty of the type the stream produces.
        # fingers crossed this also works in Python 2
        return stream.read(0)
    # find number of bytes available (this always returns a count of *bytes*)
    fcntl.ioctl(stream, termios.FIONREAD, _sock_size)
    count = _sock_size[0]
    # and read them all
    # note that count should not be zero unless the stream is closed
    # or the select call would have failed above. if we read nothing, the stream has closed.
    c = stream.read(count)
    if not c:
        # report that stream has closed
        return None
    # if we were provided with a unicode stream *and* managed to get its buffer, convert back
    if encoding:
        return c.decode(encoding, 'replace')
    return c



"""
PC-BASIC - basic.state
Support for pickling session objects

(c) 2014--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

try:
    import cPickle as pickle
except ImportError:
    import pickle

import os
import io
import sys
import zlib
import struct
import codecs
import logging
from contextlib import contextmanager

from . import VERSION
from ..compat import PY2, copyreg, stdio


# session file header
HEADER_FORMAT = '<LIIIII'
HEADER_KEYS = [
    'checksum', 'format_version', 'python_major', 'python_minor', 'pcbasic_major', 'pcbasic_minor'
]
HEADER = {
    # increment this if we change the format of the session file
    'format_version': 2,
    'python_major': sys.version_info.major,
    'python_minor': sys.version_info.minor,
    'pcbasic_major': int(VERSION.split(u'.')[0]),
    'pcbasic_minor': int(VERSION.split(u'.')[1]),
}


def unpickle_bytesio(value, pos):
    """Unpickle a file object."""
    stream = io.BytesIO(value)
    stream.seek(pos)
    return stream

def pickle_bytesio(f):
    """Pickle a BytesIO object."""
    return unpickle_bytesio, (f.getvalue(), f.tell())

def unpickle_file(name, mode, pos):
    """Unpickle a file object."""
    if name is None:
        if mode == 'rb' or (PY2 and mode == 'r'):
            return stdio.stdin.buffer
        elif mode == 'r':
            return stdio.stdin
        elif mode == 'wb' or (PY2 and mode == 'w'):
            return stdio.stdout.buffer
        elif mode == 'w':
            return stdio.stdout
    try:
        if 'w' in mode and pos > 0:
            # preserve existing contents of writable file
            with io.open(name, 'rb') as f:
                buf = f.read(pos)
            f = io.open(name, mode)
            f.write(buf)
        else:
            f = io.open(name, mode)
            if pos > 0:
                f.seek(pos)
    except IOError:
        pass
    else:
        return f
    logging.warning('Could not re-open file %s. Replacing with null file.', name)
    return io.open(os.devnull, mode)

def pickle_file(f):
    """Pickle a file object."""
    if f in (
            sys.stdout, sys.stdin,
            stdio.stdout, stdio.stdin,
            stdio.stdout.buffer, stdio.stdin.buffer
        ):
        return unpickle_file, (None, f.mode, -1)
    try:
        return unpickle_file, (f.name, f.mode, f.tell())
    except (IOError, ValueError):
        # IOError: not seekable
        # ValueError: closed
        return unpickle_file, (f.name, f.mode, -1)

# register the picklers for file and cStringIO
if PY2: # pragma: no cover
    copyreg.pickle(file, pickle_file)  # pylint: disable=undefined-variable
copyreg.pickle(io.BufferedReader, pickle_file)
copyreg.pickle(io.BufferedWriter, pickle_file)
copyreg.pickle(io.TextIOWrapper, pickle_file)
copyreg.pickle(io.BufferedRandom, pickle_file)
copyreg.pickle(io.BytesIO, pickle_bytesio)

# patch codecs.StreamReader and -Writer
if PY2: # pragma: no cover
    def patched_getstate(self):
        return vars(self)

    def patched_setstate(self, dict):
        vars(self).update(dict)

    for streamclass in (codecs.StreamReader, codecs.StreamWriter, codecs.StreamReaderWriter):
        streamclass.__getstate__ = patched_getstate
        streamclass.__setstate__ = patched_setstate


def load_session(state_file):
    """Read state from a compressed pickle."""
    if not state_file:
        raise ValueError('Session filename must not be empty')
    with open(state_file, 'rb') as in_file:
        header = in_file.read(struct.calcsize(HEADER_FORMAT))
        blob = in_file.read()
    # mask checksum to deal with different signs on Py2/Py3
    # see https://docs.python.org/3.5/library/zlib.html#zlib.crc32
    checksum = zlib.crc32(blob) & 0xffffffff
    try:
        header_dict = dict(zip(HEADER_KEYS, struct.unpack(HEADER_FORMAT, header)))
    except struct.error:
        raise ValueError('session file header corrupted')
    # check blob integrity
    if checksum != header_dict['checksum']:
        raise ValueError('session file corrupted')
    if (
            HEADER['python_major'] != header_dict['python_major']
            or HEADER['python_minor'] != header_dict['python_minor']
        ):
        raise ValueError('session file stored with different Python version')
    if (
            HEADER['pcbasic_major'] != header_dict['pcbasic_major']
            or HEADER['pcbasic_minor'] != header_dict['pcbasic_minor']
        ):
        raise ValueError('session file stored with different PC-BASIC version')
    session = pickle.loads(zlib.decompress(blob))
    return session

def save_session(obj, state_file):
    """Write state to a compressed pickle."""
    if not state_file:
        raise ValueError('Session filename must not be empty')
    blob = zlib.compress(pickle.dumps(obj, pickle.HIGHEST_PROTOCOL))
    checksum = zlib.crc32(blob) & 0xffffffff
    header_dict = dict(checksum=checksum, **HEADER)
    header = struct.pack(HEADER_FORMAT, *(header_dict[_key] for _key in HEADER_KEYS))
    with open(state_file, 'wb') as out_file:
        out_file.write(header)
        out_file.write(blob)
